from collections import OrderedDict
from ..generic.general_methods import aedt_exception_handler, generate_unique_name
from ..application.DataHandlers import dict2arg, arg2dict
import copy

defaultparametricSetup = OrderedDict({"IsEnabled": True, "ProdOptiSetupDataV2":
            OrderedDict({"SaveFields": False, "CopyMesh": False, "SolveWithCopiedMeshOnly": True}),
                                                "StartingPoint": OrderedDict(), "Sim. Setups": [],
                                                "Sweeps": OrderedDict({"SweepDefinition":
                                                                           OrderedDict({"Variable": "",
                                                                                        "Data": "",
                                                                                        "OffsetF1": False,
                                                                                        "Synchronize": 0})}),
                                                "Sweep Operations": OrderedDict(),
                                                "Goals": OrderedDict()})


defaultdxSetup = OrderedDict({"IsEnabled": True, "ProdOptiSetupDataV2":
            OrderedDict({"SaveFields": False, "CopyMesh": False, "SolveWithCopiedMeshOnly": True}),
                                                "StartingPoint": OrderedDict(), "Sim. Setups": [],
                                                "Sweeps": OrderedDict({"SweepDefinition":
                                                                           OrderedDict({"Variable": "",
                                                                                        "Data": "",
                                                                                        "OffsetF1": False,
                                                                                        "Synchronize": 0})}),
                              "Sweep Operations": OrderedDict(), "CostFunctionName": "Cost", "CostFuncNormType": "L2",
                               "CostFunctionGoals": OrderedDict(),"EmbeddedParamSetup": -1, "Goals": OrderedDict()})

defaultoptiSetup = OrderedDict({"IsEnabled": True, "ProdOptiSetupDataV2":
    OrderedDict({"SaveFields": False, "CopyMesh": False, "SolveWithCopiedMeshOnly": True}),
                                "StartingPoint": OrderedDict(), "Optimizer": "Quasi Newton",
                                "AnalysisStopOptions": OrderedDict({"StopForNumIteration": True,
                                                                    "StopForElapsTime": False,
                                                                    "StopForSlowImprovement": False,
                                                                    "StopForGrdTolerance": False,
                                                                    "MaxNumIteration": 1000,
                                                                    "MaxSolTimeInSec": 3600,
                                                                    "RelGradientTolerance": 0,
                                                                    "MinNumIteration": 10
                                                                    }),
                                "CostFuncNormType": "L2", "PriorPSetup": "", "PreSolvePSetup": True,
                                "Variables": OrderedDict(), "LCS": OrderedDict(), "Goals": OrderedDict(),
                                "Acceptable_Cost": 0, "Noise": 0.0001, "UpdateDesign": False, "UpdateIteration": 5,
                                "KeepReportAxis": True, "UpdateDesignWhenDone": True})

defaultsensitivitySetup = OrderedDict({"IsEnabled": True, "ProdOptiSetupDataV2":
    OrderedDict({"SaveFields": False, "CopyMesh": False, "SolveWithCopiedMeshOnly": True}),
                                       "StartingPoint": OrderedDict(), "MaxIterations": 10,
                                       "PriorPSetup": "", "PreSolvePSetup": True, "Variables": OrderedDict(),
                                       "LCS": OrderedDict(), "Goals": OrderedDict(), "Primary Goal": 0,
                                       "PrimaryError": 0.0001, "Perform Worst Case Analysis": False})

defaultstatisticalSetup = OrderedDict({"IsEnabled": True, "ProdOptiSetupDataV2":
    OrderedDict({"SaveFields": False, "CopyMesh": False, "SolveWithCopiedMeshOnly": True}),
                                       "StartingPoint": OrderedDict(), "MaxIterations": 50,
                                       "SeedValue": 0, "PriorPSetup": "", "Variables": OrderedDict(),
                                       "Goals": OrderedDict()})

defaultdoeSetup = OrderedDict({"IsEnabled": True, "ProdOptiSetupDataV2":
            OrderedDict({"SaveFields": False, "CopyMesh": False, "SolveWithCopiedMeshOnly": True}),
                                                "StartingPoint": OrderedDict(), "Sim. Setups": [],
                                                "CostFunctionName": "Cost",
                               "CostFuncNormType": "L2", "CostFunctionGoals": OrderedDict(), "Variables": OrderedDict(),
                               "Goals": OrderedDict(),
                               "DesignExprData": OrderedDict({"Type": "kOSF", "CCDDeignType": "kFaceCentered",
                                                              "CCDTemplateType": "kStandard",
                                                              "LHSSampleType": "kCCDSample",
                                                              "RamdomSeed": 0, "NumofSamples": 10,
                                                              "OSFDeignType": "kOSFD_MAXIMINDIST",
                                                              "MaxCydes": 10}),
                               "RespSurfaceSetupData": OrderedDict({"Type": "kGenAggr", "RefineType": "kManual"}),
                               "ResponsePoints": OrderedDict({"NumOfStrs": 0}),
                               "ManualRefinePoints": OrderedDict({"NumOfStrs": 0}),
                               "CustomVerifyPoints": OrderedDict({"NumOfStrs": 0}), "Tolerances": []})


class CommonOptimetrics(object):
    """Common Optimetrics Methods Class"""
    @property
    def omodule(self):
        """ """
        return self._parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent, name, dictinputs, optimtype):
        self._parent = parent
        self.name = name
        self.soltype = optimtype

        inputd = copy.deepcopy(dictinputs)


        if optimtype == "OptiParametric":
            self.props = inputd or defaultparametricSetup
        if optimtype == "OptiDesignExplorer":
            self.props = inputd or defaultdxSetup
        if optimtype == "OptiOptimization":
            self.props = inputd or defaultoptiSetup
        if optimtype == "OptiSensitivity":
            self.props = inputd or defaultsensitivitySetup
        if optimtype == "OptiStatistical":
            self.props = inputd or defaultstatisticalSetup
        if optimtype == "OptiDXDOE":
            self.props = inputd or defaultdoeSetup

        if inputd:
            self.props.pop('ID', None)
            self.props.pop('NextUniqueID', None)
            self.props.pop('MoveBackwards', None)
            self.props.pop('GoalSetupVersion', None)
            self.props.pop('Version', None)
            self.props.pop('SetupType', None)
            if inputd.get("Sim. Setups"):
                setups = inputd["Sim. Setups"]
                for el in setups:
                    if type(self._parent.design_properties["SolutionManager"]['ID Map']['Setup']) is list:
                        for setup in self._parent.design_properties["SolutionManager"]['ID Map']['Setup']:
                            if setup['I'] == el:
                                setups[setups.index(el)]= setup['I']
                                break
                    else:
                        if self._parent.design_properties["SolutionManager"]['ID Map']['Setup']['I'] == el:
                            setups[setups.index(el)] = self._parent.design_properties["SolutionManager"]['ID Map']['Setup']['N']
                            break
            if inputd.get("Goals", None):
                oparams = self.omodule.GetChildObject(self.name).GetCalculationInfo()
                oparam = [i for i in oparams[0]]
                calculation = ["NAME:Goal"]
                calculation.extend(oparam)
                arg1 = OrderedDict()
                arg2dict(calculation, arg1)
                self.props['Goals'] = arg1

    @aedt_exception_handler
    def update(self, update_dictionary=None):
        """update a Setup based on stored properties. a dictionary can be provided as argument

        Parameters
        ----------
        update_dictionary :
            optional dictionary argument (Default value = None)

        Returns
        -------
        type
            Bool

        """
        if update_dictionary:
            for el in update_dictionary:
                self.props[el]=update_dictionary[el]

        arg = ["NAME:" + self.name]
        dict2arg(self.props, arg)

        self.omodule.EditSetup(self.name, arg)
        return True

    @aedt_exception_handler
    def create(self):
        """update a Setup based on stored properties.
        
        :return: Bool

        Parameters
        ----------

        Returns
        -------

        """
        arg = ["NAME:" + self.name]
        dict2arg(self.props, arg)
        self.omodule.InsertSetup(self.soltype, arg)
        return True

    @aedt_exception_handler
    def _add_calculation(self, reporttype, solution=None, domain="Sweep", calculation="", calculation_type="d",
                         calculation_value="", calculation_name=None):
        """

        Parameters
        ----------
        reporttype :
            
        solution :
             (Default value = None)
        domain :
             (Default value = "Sweep")
        calculation :
             (Default value = "")
        calculation_type :
             (Default value = "d")
        calculation_value :
             (Default value = "")
        calculation_name :
             (Default value = None)

        Returns
        -------

        """
        sweepdefinition = OrderedDict()
        sweepdefinition["ReportType"] = reporttype
        if not solution:
            solution = self._parent.nominal_sweep

        sweepdefinition["Solution"] = solution
        sweepdefinition["SimValueContext"] = OrderedDict({"Domain": domain})
        sweepdefinition["Calculation"] = calculation
        if calculation_name:
            sweepdefinition["Name"] = calculation_name
        else:
            sweepdefinition["Name"] = generate_unique_name(calculation)
        if domain == "Sweep":
            var = "Freq"
        else:
            var ="Time"
        sweepdefinition["Ranges"] = OrderedDict(
            {"Range": ["Var:=", var, "Type:=", calculation_type, "DiscreteValues:=", calculation_value]})
        if "Goal" in self.props["Goals"]:
            if type(self.props["Goals"]["Goal"]) is not list:
                self.props["Goals"]["Goal"] = [self.props["Goals"]["Goal"], sweepdefinition]
            else:
                self.props["Goals"]["Goal"].append(sweepdefinition)
        else:
            self.props["Goals"]["Goal"] = sweepdefinition

        return self.update()

    @aedt_exception_handler
    def _add_goal(self, optigoalname,reporttype, solution=None, domain="Sweep", calculation="", calculation_type="discrete",
                  calc_val1="", calc_val2="", condition="==", goal_value=1, goal_weight=1, goal_name=None):
        """

        Parameters
        ----------
        optigoalname :
            
        reporttype :
            
        solution :
             (Default value = None)
        domain :
             (Default value = "Sweep")
        calculation :
             (Default value = "")
        calculation_type :
             (Default value = "discrete")
        calc_val1 :
             (Default value = "")
        calc_val2 :
             (Default value = "")
        condition :
             (Default value = "==")
        goal_value :
             (Default value = 1)
        goal_weight :
             (Default value = 1)
        goal_name :
             (Default value = None)

        Returns
        -------

        """
        sweepdefinition = OrderedDict()
        sweepdefinition["ReportType"] = reporttype
        if not solution:
            solution = self._parent.nominal_sweep
        sweepdefinition["Solution"] = solution
        sweepdefinition["SimValueContext"] = OrderedDict({"Domain": domain})
        sweepdefinition["Calculation"] = calculation
        if goal_name:
            sweepdefinition["Name"] = goal_name
        else:
            sweepdefinition["Name"] = generate_unique_name(calculation)
        if domain == "Sweep":
            var = "Freq"
        else:
            var ="Time"
        if calculation_type =="discrete":
            if type(calc_val1) is list:
                dr = ",".join(calc_val1)
            else:
                dr = calc_val1
            sweepdefinition["Ranges"] = OrderedDict(
                {"Range": ["Var:=", var, "Type:=", "d", "DiscreteValues:=", dr]})
        else:
            sweepdefinition["Ranges"] = OrderedDict({"Range": ["Var:=", var, "Type:=", calculation_type,
                                                               "Start:=", calc_val1, "Stop:=", calc_val2,
                                                               "DiscreteValues:=", ""]})
        sweepdefinition["Condition"] = condition
        sweepdefinition["GoalValue"] = OrderedDict(
            {"GoalValueType": "Independent", "Format": "Real/Imag", "bG": ["v:=", "[{};]".format(goal_value)]})
        sweepdefinition["Weight"] = "[{};]".format(goal_weight)
        if "Goal" in self.props[optigoalname]:
            if type(self.props[optigoalname]["Goal"]) is not list:
                self.props[optigoalname]["Goal"] = [self.props[optigoalname]["Goal"], sweepdefinition]
            else:
                self.props[optigoalname]["Goal"].append(sweepdefinition)
        else:
            self.props[optigoalname]["Goal"] = sweepdefinition
        return self.update()


class DXSetups(object):
    """Design XPlorer Setups Class"""

    class Setup(CommonOptimetrics, object):
        """Setup Class"""
        def __init__(self, parent, name, dictinputs=None):
            CommonOptimetrics.__init__(self, parent, name, dictinputs=dictinputs, optimtype="OptiDesignExplorer")
            pass

        @aedt_exception_handler
        def add_calculation(self, calculation="", calculation_value="", reporttype="Modal Solution Data", solution=None, domain="Sweep",  calculation_name=None):
            """Add Calculation to DX Setup

            Parameters
            ----------
            reporttype :
                Name of the Report to take the Calculation (Default value = "Modal Solution Data")
            solution :
                solution type. if none, default solution will be taken
            domain :
                domain type. if none sweep will be taken (Default value = "Sweep")
            calculation :
                calculation value eg. dB(S(1,1,)) (Default value = "")
            calculation_value :
                calculation value. eg. if sweep "1GHz) (Default value = "")
            calculation_name :
                 (Default value = None)

            Returns
            -------
            type
                bool

            """
            return self._add_calculation(reporttype=reporttype, solution=solution, domain=domain, calculation_type="d",
                                         calculation=calculation, calculation_value=calculation_value, calculation_name=calculation_name)

        @aedt_exception_handler
        def add_goal(self, calculation="", calculation_value="", calculation_type="discrete", calculation_stop="",
                     reporttype="Modal Solution Data", solution=None, domain="Sweep", goal_name=None, goal_value=1, goal_weight=1, condition="=="):
            """Add Goald to DX Setup

            Parameters
            ----------
            reporttype :
                Name of the Report to take the Calculation (Default value = "Modal Solution Data")
            solution :
                solution type. if none, default solution will be taken
            domain :
                domain type. if none sweep will be taken (Default value = "Sweep")
            calculation :
                calculation value eg. dB(S(1,1,)) (Default value = "")
            calculation_type :
                discrete or range (Default value = "discrete")
            calculation_value :
                calculation value. eg. if sweep "1GHz). if discrete, it is discrete value or list, if range, it is start value (Default value = "")
            calculation_stop :
                calculation value. eg. if sweep "1GHz). if range it is the stop value (Default value = "")
            goal_name :
                 (Default value = None)
            goal_value :
                 (Default value = 1)
            goal_weight :
                 (Default value = 1)
            condition :
                 (Default value = "==")

            Returns
            -------
            type
                bool

            """
            return self._add_goal(optigoalname="CostFunctionGoals", reporttype=reporttype, solution=solution,
                                  domain=domain, calculation_type=calculation_type, calculation=calculation,
                                  calc_val1=calculation_value, calc_val2=calculation_stop, goal_name=goal_name,
                                  goal_weight=goal_weight, goal_value=goal_value, condition=condition)
    @property
    def parent(self):
        """ """
        return self._parent
    @property
    def optimodule(self):
        """ """
        return self.parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent):
        self._parent = parent
        self.setups = []
        if self._parent.design_properties:
            try:
                setups_data = self._parent.design_properties["Optimetrics"]["OptimetricsSetups"]
                for data in setups_data:
                    if type(setups_data[data]) is OrderedDict and setups_data[data]['SetupType'] == "OptiDesignExplorer":
                        self.setups.append(self.Setup(parent, data, setups_data[data]))
            except:
                pass

    @aedt_exception_handler
    def add_dx_setup(self,variables_to_include, defaults_var_values=None, setupname=None, parametricname=None):
        """Add a DesignXplorer Basic Setup. User can customize all DX Options after the creation

        Parameters
        ----------
        variables_to_include :
            list of variables to include in DesignXplorer
        defaults_var_values :
            list of default variable values
        setupname :
            Optional Setup name. if none, the default analysis setup will be used
        parametricname :
            optional Parametric Name (Default value = None)

        Returns
        -------
        type
            Optimetrics Object Class

        """
        if not setupname:
            setupname = [self._parent.analysis_setup]
        elif type(setupname) is not list:
            setupname = [setupname]
        if not parametricname:
            parametricname = generate_unique_name("DesignXplorer")
        setup = self.Setup(self._parent, parametricname)
        setup.props["Sim. Setups"] = setupname
        setup.props['Sweeps'] = []
        if not defaults_var_values:
            for v in variables_to_include:
                sweepdefinition = OrderedDict()
                sweepdefinition["Variable"] = v
                if "$" in v:
                    sweepdefinition["Data"] = self._parent.oproject.GetVariableValue(v)
                else:
                    sweepdefinition["Data"] = self._parent.odesign.GetVariableValue(v)
                sweepdefinition["OffsetF1"] = False
                sweepdefinition["Synchronize"] = 0
        else:
            for v,vv in zip(variables_to_include, defaults_var_values):
                sweepdefinition = OrderedDict()
                sweepdefinition["Variable"] = v
                sweepdefinition["Data"] = vv
                sweepdefinition["OffsetF1"] = False
                sweepdefinition["Synchronize"] = 0
                setup.props['Sweeps'].append(sweepdefinition)
                setup.props["StartingPoint"][v] = vv
        setup.create()
        self.setups.append(setup)
        return setup


class ParametericsSetups(object):
    """Parametric Setup Class"""
    class Setup(CommonOptimetrics, object):
        """Setup Class"""
        def __init__(self, parent, name, dictinputs=None):
            CommonOptimetrics.__init__(self, parent, name, dictinputs=dictinputs, optimtype="OptiParametric")
            pass

        @aedt_exception_handler
        def add_variation(self, sweep_var, datarange):
            """Add a Variation to an existing Parametric Setup

            Parameters
            ----------
            sweep_var :
                Variable name
            datarange :
                Data Range

            Returns
            -------
            type
                Bool

            """
            if type(self.props["Sweeps"]["SweepDefinition"]) is not list:
                self.props["Sweeps"]["SweepDefinition"] = [self.props["Sweeps"]["SweepDefinition"]]
            sweepdefinition = OrderedDict()
            sweepdefinition["Variable"] = sweep_var
            sweepdefinition["Data"] = datarange
            sweepdefinition["OffsetF1"] = False
            sweepdefinition["Synchronize"] = 0
            self.props["Sweeps"]["SweepDefinition"].append(sweepdefinition)
            self.update()
            return True

        @aedt_exception_handler
        def add_calculation(self, calculation="", calculation_value="", reporttype="Modal Solution Data", solution=None,
                            domain="Sweep", calculation_name=None):
            """Add Calculation to Parametric Setup

            Parameters
            ----------
            reporttype :
                Name of the Report to take the Calculation (Default value = "Modal Solution Data")
            solution :
                solution type. if none, default solution will be taken
            domain :
                domain type. if none sweep will be taken (Default value = "Sweep")
            calculation :
                calculation value eg. dB(S(1,1,)) (Default value = "")
            calculation_value :
                calculation value. eg. if sweep "1GHz) (Default value = "")
            calculation_name :
                 (Default value = None)

            Returns
            -------
            type
                bool

            """
            return self._add_calculation(reporttype=reporttype, solution=solution, domain=domain, calculation_type="d",
                                         calculation=calculation, calculation_value=calculation_value, calculation_name=calculation_name)



    @property
    def parent(self):
        """ """
        return self._parent

    @property
    def optimodule(self):
        """ """
        return self.parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent):
        self._parent = parent
        self.setups = []
        if self._parent.design_properties:
            try:
                setups_data = self._parent.design_properties["Optimetrics"]["OptimetricsSetups"]

                for data in setups_data:
                    if type(setups_data[data]) is OrderedDict and setups_data[data]['SetupType'] == "OptiParametric":
                        self.setups.append(self.Setup(parent, data, setups_data[data]))
            except:
                pass

    @aedt_exception_handler
    def add_parametric_setup(self, sweep_var, datarange, setupname=None, parametricname=None):
        """Add a Parametric Basic Setup. User can customize all Options after the creation

        Parameters
        ----------
        sweep_var :
            Variable name
        datarange :
            Data Range
        setupname :
            Optional Setup name. if none, the default analysis setup will be used
        parametricname :
            optional Parametric Name (Default value = None)

        Returns
        -------
        type
            Optimetrics Object Class

        """
        if not setupname:
            setupname = [self._parent.analysis_setup]
        elif type(setupname) is not list:
            setupname = [setupname]
        if not parametricname:
            parametricname = generate_unique_name("Parametric")
        setup = self.Setup(self._parent, parametricname)
        setup.props["Sim. Setups"] = setupname
        sweepdefinition = OrderedDict()
        sweepdefinition["Variable"] = sweep_var
        sweepdefinition["Data"] = datarange
        sweepdefinition["OffsetF1"] = False
        sweepdefinition["Synchronize"] = 0
        setup.props["Sweeps"]["SweepDefinition"] = sweepdefinition
        setup.create()
        self.setups.append(setup)
        return setup


class SensitivitySetups(object):
    """Sensitivity Class"""
    class Setup(CommonOptimetrics, object):
        """Setup Class"""
        def __init__(self, parent, name, dictinputs=None):
            CommonOptimetrics.__init__(self, parent, name, dictinputs=dictinputs, optimtype="OptiSensitivity")
            pass

        @aedt_exception_handler
        def add_calculation(self, calculation="", calculation_value="", reporttype="Modal Solution Data", solution=None,
                            domain="Sweep", calculation_name=None):
            """Add Calculation to Sensitivity Setup

            Parameters
            ----------
            reporttype :
                Name of the Report to take the Calculation (Default value = "Modal Solution Data")
            solution :
                solution type. if none, default solution will be taken
            domain :
                domain type. if none sweep will be taken (Default value = "Sweep")
            calculation :
                calculation value eg. dB(S(1,1,)) (Default value = "")
            calculation_value :
                calculation value. eg. if sweep "1GHz) (Default value = "")
            calculation_name :
                 (Default value = None)

            Returns
            -------
            type
                bool

            """
            return self._add_calculation(reporttype=reporttype, solution=solution, domain=domain, calculation_type="d",
                                         calculation=calculation, calculation_value=calculation_value, calculation_name=calculation_name)

    @property
    def parent(self):
        """ """
        return self._parent
    @property
    def optimodule(self):
        """ """
        return self.parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent):
        self._parent = parent
        self.setups = []
        if self._parent.design_properties:
            try:
                setups_data = self._parent.design_properties["Optimetrics"]["OptimetricsSetups"]
                for data in setups_data:
                    if type(setups_data[data]) is OrderedDict and setups_data[data]['SetupType'] == "OptiSensitivity":
                        self.setups.append(self.Setup(parent, data, setups_data[data]))
            except:
                pass

    @aedt_exception_handler
    def add_sensitivity(self, calculation, calculation_value, calculation_type="Freq",
                        reporttype="Modal Solution Data", domain="Sweep", solution=None, parametricname=None):
        """Add a Sensitivity Basic Setup. User can customize all Options after the creation

        Parameters
        ----------
        calculation :
            name of the Calculation
        calculation_value :
            Variation Value (eg. "1GHz")
        calculation_type :
            Variation type. Default Freq
        reporttype :
            Report Type. Default "Modal Solution Data"
        domain :
            Default "Sweep"
        solution :
            Solution name. If nothing, nominal sweep will be used (Default value = None)
        parametricname :
            Name of the sensititivy analsysis. if nothing, a default name will be given

        Returns
        -------
        type
            Sensitivity object

        """
        if not parametricname:
            parametricname = generate_unique_name("Sensitivity")
        setup = self.Setup(self._parent, parametricname)
        sweepdefinition = OrderedDict()
        sweepdefinition["ReportType"] = reporttype
        if not solution:
            solution = self._parent.nominal_sweep
        sweepdefinition["Solution"] = solution
        sweepdefinition["SimValueContext"] = OrderedDict({"Domain": domain})
        sweepdefinition["Calculation"] = calculation
        sweepdefinition["Name"] = calculation
        sweepdefinition["Ranges"] = OrderedDict(
            {"Range": ["Var:=", calculation_type, "Type:=", "d", "DiscreteValues:=", calculation_value]})
        setup.props["Goals"]["Goal"] = sweepdefinition
        setup.create()
        self.setups.append(setup)
        return setup


class StatisticalSetups(object):
    """Statistical Class"""
    class Setup(CommonOptimetrics, object):
        """Setup Class"""
        def __init__(self, parent, name, dictinputs=None):
            CommonOptimetrics.__init__(self, parent, name, dictinputs=dictinputs, optimtype="OptiStatistical")
            pass

        @aedt_exception_handler
        def add_calculation(self, calculation="", calculation_value="", reporttype="Modal Solution Data", solution=None, domain="Sweep",  calculation_name=None):
            """Add Calculation to Statistical Setup

            Parameters
            ----------
            reporttype :
                Name of the Report to take the Calculation (Default value = "Modal Solution Data")
            solution :
                solution type. if none, default solution will be taken
            domain :
                domain type. if none sweep will be taken (Default value = "Sweep")
            calculation :
                calculation value eg. dB(S(1,1,)) (Default value = "")
            calculation_value :
                calculation value. eg. if sweep "1GHz) (Default value = "")
            calculation_name :
                 (Default value = None)

            Returns
            -------
            type
                bool

            """
            return self._add_calculation(reporttype=reporttype, solution=solution, domain=domain, calculation_type="d",
                                         calculation=calculation, calculation_value=calculation_value, calculation_name=calculation_name)

    @property
    def parent(self):
        """ """
        return self._parent
    @property
    def optimodule(self):
        """ """
        return self.parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent):
        self._parent = parent
        self.setups = []
        if self._parent.design_properties:
            try:
                setups_data = self._parent.design_properties["Optimetrics"]["OptimetricsSetups"]
                for data in setups_data:
                    if type(setups_data[data]) is OrderedDict and setups_data[data]['SetupType'] == "OptiStatistical":
                        self.setups.append(self.Setup(parent, data, setups_data[data]))
            except:
                pass

    @aedt_exception_handler
    def add_statistical(self, calculation_name, calc_variation_value, calculation_type="Freq",
                        reporttype="Modal Solution Data", domain="Sweep", solution=None, parametricname=None):
        """Add a Sensitivity Basic Setup. User can customize all Options after the creation

        Parameters
        ----------
        calculation_name :
            name of the Calculation
        calc_variation_value :
            Variation Value (eg. "1GHz")
        calculation_type :
            Variation type. Default Freq
        reporttype :
            Report Type. Default "Modal Solution Data"
        domain :
            Default "Sweep"
        solution :
            Solution name. If nothing, nominal sweep will be used (Default value = None)
        parametricname :
            Name of the sensititivy analsysis. if nothing, a default name will be given

        Returns
        -------
        type
            Statistical object

        """
        if not parametricname:
            parametricname = generate_unique_name("Statistical")
        setup = self.Setup(self._parent, parametricname)
        sweepdefinition = OrderedDict()
        sweepdefinition["ReportType"] = reporttype
        if not solution:
            solution = self._parent.nominal_sweep

        sweepdefinition["Solution"] = solution
        sweepdefinition["SimValueContext"] = OrderedDict({"Domain": domain})
        sweepdefinition["Calculation"] = calculation_name
        sweepdefinition["Name"] = calculation_name
        sweepdefinition["Ranges"] = OrderedDict({"Range": ["Var:=", calculation_type,	"Type:=", "d", "DiscreteValues:=", calc_variation_value]})
        setup.props["Goals"]["Goal"] = sweepdefinition
        setup.create()
        self.setups.append(setup)
        return setup


class DOESetups(object):
    """DOE Class"""
    class Setup(CommonOptimetrics, object):
        """Setup Class"""
        def __init__(self, parent, name, dictinputs=None):
            CommonOptimetrics.__init__(self, parent, name, dictinputs=dictinputs, optimtype="OptiDXDOE")
            pass
        @aedt_exception_handler
        def add_calculation(self, calculation="", calculation_value="", reporttype="Modal Solution Data", solution=None,
                            domain="Sweep", calculation_name=None):
            """Add Calculation to DOE Setup

            Parameters
            ----------
            reporttype :
                Name of the Report to take the Calculation (Default value = "Modal Solution Data")
            solution :
                solution type. if none, default solution will be taken
            domain :
                domain type. if none sweep will be taken (Default value = "Sweep")
            calculation :
                calculation value eg. dB(S(1,1,)) (Default value = "")
            calculation_value :
                calculation value. eg. if sweep "1GHz) (Default value = "")
            calculation_name :
                 (Default value = None)

            Returns
            -------
            type
                bool

            """
            return self._add_calculation(reporttype=reporttype, solution=solution, domain=domain, calculation_type="d",
                                         calculation=calculation, calculation_value=calculation_value, calculation_name=calculation_name)

        @aedt_exception_handler
        def add_goal(self, calculation="", calculation_value="", calculation_type="discrete",
                     calculation_stop="", reporttype="Modal Solution Data", solution=None, domain="Sweep", goal_name=None, goal_value=1, goal_weight=1, condition="=="):
            """Add Goald to DOE Setup

            Parameters
            ----------
            reporttype :
                Name of the Report to take the Calculation (Default value = "Modal Solution Data")
            solution :
                solution type. if none, default solution will be taken
            domain :
                domain type. if none sweep will be taken (Default value = "Sweep")
            calculation :
                calculation value eg. dB(S(1,1,)) (Default value = "")
            calculation_type :
                discrete or range (Default value = "discrete")
            calculation_value :
                calculation value. eg. if sweep "1GHz). if discrete, it is discrete value or list, if range, it is start value (Default value = "")
            calculation_stop :
                calculation value. eg. if sweep "1GHz). if range it is the stop value (Default value = "")
            goal_name :
                 (Default value = None)
            goal_value :
                 (Default value = 1)
            goal_weight :
                 (Default value = 1)
            condition :
                 (Default value = "==")

            Returns
            -------
            type
                bool

            """
            return self._add_goal(optigoalname="CostFunctionGoals", reporttype=reporttype, solution=solution,
                                  domain=domain, calculation_type=calculation_type, calculation=calculation,
                                  calc_val1=calculation_value, calc_val2=calculation_stop, goal_name=goal_name,
                                  goal_weight=goal_weight,
                                  goal_value=goal_value, condition=condition)

    @property
    def parent(self):
        """ """
        return self._parent
    @property
    def optimodule(self):
        """ """
        return self.parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent):
        self._parent = parent
        self.setups = []
        if self._parent.design_properties:
            try:
                setups_data = self._parent.design_properties["Optimetrics"]["OptimetricsSetups"]
                for data in setups_data:
                    if type(setups_data[data]) is OrderedDict and setups_data[data]['SetupType'] == "OptiDXDOE":
                        self.setups.append(self.Setup(parent, data, setups_data[data]))
            except:
                pass

    @aedt_exception_handler
    def add_doe(self, calculation, calculation_value, calculation_type="Freq", reporttype="Modal Solution Data",
                domain="Sweep", condition="<=", goal_value=1, goal_weight=1, solution=None, parametricname=None):
        """Add a DOE Basic Setup. User can customize all Options after the creation

        Parameters
        ----------
        calculation :
            name of the Calculation
        calculation_value :
            Variation Value (eg. "1GHz")
        calculation_type :
            Variation type. Default Freq
        reporttype :
            Report Type. Default "Modal Solution Data"
        domain :
            Default "Sweep"
        condition :
            goal condition. Default "<="
        goal_value :
            goal value to meet condition. Default 1
        goal_weight :
            goal weight. Default 1
        solution :
            Solution name. If nothing, nominal sweep will be used (Default value = None)
        parametricname :
            Name of the sensititivy analsysis. if nothing, a default name will be given

        Returns
        -------
        type
            DOE object

        """
        if not solution:
            solution = self._parent.nominal_sweep
        setupname = [solution.split(" ")[0]]
        if not parametricname:
            parametricname = generate_unique_name("DesignOfExp")
        setup = self.Setup(self._parent, parametricname)
        setup.props["Sim. Setups"] = setupname
        sweepdefinition = OrderedDict()
        sweepdefinition["ReportType"] = reporttype

        sweepdefinition["Solution"] = solution
        sweepdefinition["SimValueContext"] = OrderedDict({"Domain": domain})
        sweepdefinition["Calculation"] = calculation
        sweepdefinition["Name"] = calculation
        sweepdefinition["Ranges"] = OrderedDict(
            {"Range": ["Var:=", calculation_type, "Type:=", "d", "DiscreteValues:=", calculation_value]})
        sweepdefinition["Condition"] = condition
        sweepdefinition["GoalValue"] = OrderedDict(
            {"GoalValueType": "Independent", "Format": "Real/Imag", "bG": ["v:=", "[{};]".format(goal_value)]})
        sweepdefinition["Weight"] = "[{};]".format(goal_weight)
        setup.props["CostFunctionGoals"]["Goal"] = sweepdefinition
        setup.create()
        self.setups.append(setup)
        return setup


class OptimizationSetups(object):
    """Optimization Class"""
    class Setup(CommonOptimetrics, object):
        """Setup Class"""
        def __init__(self, parent, name, dictinputs=None):
            CommonOptimetrics.__init__(self, parent, name, dictinputs=dictinputs, optimtype="OptiOptimization")
            pass

        @aedt_exception_handler
        def add_goal(self, calculation="", calculation_value="", calculation_type="discrete", calculation_stop="",
                     reporttype="Modal Solution Data", solution=None, domain="Sweep", goal_name=None, goal_value=1,
                     goal_weight=1, condition="=="):
            """Add Calculation to Optimetrics Setup

            Parameters
            ----------
            reporttype :
                Name of the Report to take the Calculation (Default value = "Modal Solution Data")
            solution :
                solution type. if none, default solution will be taken
            domain :
                domain type. if none sweep will be taken (Default value = "Sweep")
            calculation :
                calculation value eg. dB(S(1,1,)) (Default value = "")
            calculation_type :
                discrete or range (Default value = "discrete")
            calculation_value :
                calculation value. eg. if sweep "1GHz). if discrete, it is discrete value or list, if range, it is start value (Default value = "")
            calculation_stop :
                calculation value. eg. if sweep "1GHz). if range it is the stop value (Default value = "")
            goal_value :
                goal Value. Default 1
            goal_weight :
                Goal Weight. Default 1
            condition :
                Condition. Default "=="
            goal_name :
                 (Default value = None)

            Returns
            -------
            type
                bool

            """
            return self._add_goal(optigoalname="Goals", reporttype=reporttype, solution=solution, domain=domain,
                                  calculation_type=calculation_type, calculation=calculation,
                                  calc_val1=calculation_value, calc_val2=calculation_stop, goal_name=goal_name, goal_value=goal_value, goal_weight=goal_weight, condition=condition)

    @property
    def parent(self):
        """ """
        return self._parent
    @property
    def optimodule(self):
        """ """
        return self.parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent):
        self._parent = parent
        self.setups = []
        if self._parent.design_properties:
            try:
                setups_data = self._parent.design_properties["Optimetrics"]["OptimetricsSetups"]
                for data in setups_data:
                    if type(setups_data[data]) is OrderedDict and setups_data[data]['SetupType'] == "OptiOptimization":
                        self.setups.append(self.Setup(parent, data, setups_data[data]))
            except:
                pass



    @aedt_exception_handler
    def add_optimization(self, calculation, calculation_value, calculation_type="Freq",
                         reporttype="Modal Solution Data", domain="Sweep", condition="<=", goal_value=1, goal_weight=1,
                         solution=None, parametricname=None):
        """Add a Optimization Basic Setup. User can customize all Options after the creation

        Parameters
        ----------
        calculation :
            name of the Calculation
        calculation_value :
            Variation Value (eg. "1GHz")
        calculation_type :
            Variation type. Default Freq
        reporttype :
            Report Type. Default "Modal Solution Data"
        domain :
            Default "Sweep"
        condition :
            goal condition. Default "<="
        goal_value :
            goal value to meet condition. Default 1
        goal_weight :
            goal weight. Default 1
        solution :
            Solution name. If nothing, nominal sweep will be used (Default value = None)
        parametricname :
            Name of the sensititivy analsysis. if nothing, a default name will be given

        Returns
        -------
        type
            Optimization object

        """
        if not parametricname:
            parametricname = generate_unique_name("Optimization")
        setup = self.Setup(self._parent, parametricname)
        sweepdefinition = OrderedDict()
        sweepdefinition["ReportType"] = reporttype
        if not solution:
            solution = self._parent.nominal_sweep
        sweepdefinition["Solution"] = solution
        sweepdefinition["SimValueContext"] = OrderedDict({"Domain": domain})
        sweepdefinition["Calculation"] = calculation
        sweepdefinition["Name"] = calculation
        sweepdefinition["Ranges"] = OrderedDict({"Range": ["Var:=", calculation_type,	"Type:=", "d", "DiscreteValues:=", calculation_value]})
        sweepdefinition["Condition"] = condition
        sweepdefinition["GoalValue"] = OrderedDict(
            {"GoalValueType": "Independent", "Format": "Real/Imag", "bG": ["v:=", "[{};]".format(goal_value)]})
        sweepdefinition["Weight"] = "[{};]".format(goal_weight)

        setup.props["Goals"]["Goal"] = sweepdefinition
        setup.create()
        self.setups.append(setup)
        return setup



from __future__ import absolute_import
import warnings
from .general import *
from ..generic.general_methods import get_filename_without_extension, generate_unique_name

try:
    import clr
    from System import Convert, String
    from System import Double, Array
    from System.Collections.Generic import List
except ImportError:
    warnings.warn('This module requires pythonnet.')


class EdbNets(object):
    """EdbNets class"""

    def __init__(self, parent):
        self.parent = parent

    @property
    def builder(self):
        """ """
        return self.parent.builder

    @property
    def edb(self):
        """ """
        return self.parent.edb

    @property
    def edb_value(self):
        """ """
        return self.parent.edb_value

    @property
    def active_layout(self):
        """ """
        return self.parent.active_layout

    @property
    def cell(self):
        """ """
        return self.parent.cell

    @property
    def db(self):
        """ """
        return self.parent.db

    @property
    def padstack_methods(self):
        """ """
        return self.parent.edblib.Layout.PadStackMethods

    @property
    def messenger(self):
        """ """
        return self.parent._messenger

    @property
    def nets_methods(self):
        """ """
        return self.parent.edblib.Layout.NetsMethods

    @property
    def nets(self):
        """Nets.
        
        Returns
        -------
        dict
            Dictionary of nets.
        """
        if self.builder:
            return convert_netdict_to_pydict(self.nets_methods.GetNetDict(self.builder))

    @property
    def signal_nets(self):
        """Signal nets.
        
        Return
        ------
        dict
            Dictionary of signal nets.
        """
        if self.builder:
            return convert_netdict_to_pydict(self.nets_methods.GetSignalNetDict(self.builder))

    @property
    def power_nets(self):
        """Power nets.
        
        Returns
        -------
        dict
            Dictionary of power nets.
        """
        if self.builder:
            return convert_netdict_to_pydict(self.nets_methods.GetPowerNetDict(self.builder))

    @aedt_exception_handler
    def is_power_gound_net(self, netname_list):
        """Determine if one of the  nets in a list is power or ground.

        Parameters
        ----------
        netname_list : list
            List of net names.

        Returns
        -------
        bool
            ``True`` when one of the net names is ``"power"`` or ``"ground"``, ``False`` otherwise.
        """
        if self.builder:
            return self.nets_methods.IsPowerGroundNetInList(self.builder, netname_list)

    def get_dcconnected_net_list(self, ground_nets=["GND"]):
        """Retrieve the nets connected to DC through inductors.
        
        .. note::
           Only inductors are considered.

        Parameters
        ----------
        ground_nets : list, optional
            List of ground nets. The default is ``["GND"]``.

        Returns
        -------
        list
            List of nets connected to DC through inductors. 
        """
        temp_list = []
        for refdes, comp_obj in self.parent.core_components.inductors.items():

            numpins = comp_obj.numpins

            if numpins == 2:
                nets = comp_obj.nets
                if not set(nets).intersection(set(ground_nets)):
                    temp_list.append(set(nets))
                else:
                    pass

        dcconnected_net_list = []

        while not not temp_list:
            s = temp_list.pop(0)
            interseciton_flag = False
            for i in temp_list:
                if not not s.intersection(i):
                    i.update(s)
                    interseciton_flag = True

            if not interseciton_flag:
                dcconnected_net_list.append(s)

        return dcconnected_net_list

    def get_powertree(self, power_net_name, ground_nets):
        """Retrieve the power tree.

        Parameters
        ----------
        power_net_name : str
            Name of the power net.
        ground_nets :
            

        Returns
        -------

        """
        flag_in_ng = False
        net_group = []
        for ng in self.get_dcconnected_net_list(ground_nets):
            if power_net_name in ng:
                flag_in_ng = True
                net_group.extend(ng)
                break

        if not flag_in_ng:
            net_group.append(power_net_name)

        df_list = []
        rats = self.parent.core_components.get_rats()
        for net in net_group:
            for el in rats:
                if net in el['net_name']:
                    i = 0
                    for n in el['net_name']:
                        if n == net:
                            df = [el['refdes'][i], el['pin_name'][i],net ]
                            df_list.append(df)
                        i += 1

        component_type = []
        for el in df_list:
            refdes = el[0]
            comp_type = self.parent.core_components._cmp[refdes].type
            component_type.append(comp_type)
            el.append(comp_type)
        return df_list, net_group

    @aedt_exception_handler
    def get_net_by_name(self, net_name):
        edb_net = self.parent.edb.Cell.Net.FindByName(self.active_layout, net_name)
        if edb_net is not None:
            return edb_net

    @aedt_exception_handler
    def delete_nets(self, netlist):
        """Delete one or more nets from EDB.
       
        Parameters
        ----------
        netlist : str or list
            One or more nets to delete.

        Returns
        -------
        list
            List of nets that were deleted.

        Examples
        --------
        
        >>> deleted_nets = edb_core.core_nets.delete_nets(["Net1","Net2"])
        """
        if type(netlist) is str:
            netlist=[netlist]
        nets_deleted =[]
        for net in netlist:
            try:
                edb_net = self.edb.Cell.Net.FindByName(self.active_layout, net)
                if edb_net is not None:
                    edb_net.Delete()
                    nets_deleted.append(net)
                    self.parent._messenger.add_info_message("Net {} Deleted".format(net))
            except:
                pass

        return nets_deleted

    def find_or_create_net(self, net_name=''):
        """Find or create the net with the given name in the layout.

        Parameters
        ----------
        net_name : str, optional
            Name of the net to find or create. The default is ``""``.

        Returns
        -------
        object
            Net Object
        """

        net = self.edb.Cell.Net.FindByName(self.active_layout, net_name)
        if net.IsNull():
            net = self.edb.Cell.Net.Create(self.active_layout, net_name)
        return net





"""
Variables Library Class
----------------------------------------------------------------


Description
==================================================

This class contains all the functionalities to create and edit design and project variable in a very easy way in all the 3D Tools

:Example:

>>> from pyaedt import Hfss
>>> hfss = Hfss()
>>> hfss["$d"] = "5mm"
>>> hfss["d"] = "5mm"
>>> hfss["postd"] = "1W"


================

"""
from __future__ import absolute_import
import math
import re
import numbers
import os
from .. import aedt_exception_handler

@aedt_exception_handler
def dB(x, inverse=True):
    if inverse:
        return 20*math.log10(x)
    else:
        return math.pow(10, x/20.0)


@aedt_exception_handler
def fah2kel(val, inverse=True):
    if inverse:
        return (val - 273.15) * 9 / 5 + 32
    else:
        return (val - 32) * 5 / 9 + 273.15


@aedt_exception_handler
def cel2kel(val, inverse=True):
    if inverse:
        return val - 273.15
    else:
        return val + 273.15


@aedt_exception_handler
def unit_system(units):
    """ Return the name of the unit system associated with the given unit string. Return False if the units are
        not known as defined in AEDT_units """
    found = False
    for unit_type, unit_list in AEDT_units.items():
        for test_unit in unit_list:
            if test_unit == units:
                found = True
                break
        if found:
            return unit_type
    return False

#TODO Add additional units
rad2deg = 180.0 / math.pi
deg2rad = math.pi / 180
hour2sec = 3600.0
min2sec = 60.0
sec2min = 1 / 60.0
sec2hour = 1 / 3600.0
inv2pi = 0.5/math.pi
v2pi = 2.0 * math.pi
m2in = 0.0254
m2miles = 1609.344051499

# Value in new units equals value in old units * scaling vactor of new units
# e.g. value_in_deg Variable("1rad") = 1 * rad2deg
AEDT_units = {
    'AngularSpeed': {'deg_per_hr': hour2sec * deg2rad, 'rad_per_hr': hour2sec, 'deg_per_min': min2sec * deg2rad,
                     'rad_per_min': min2sec, 'deg_per_sec': deg2rad, 'rad_per_sec': 1.0, 'rev_per_sec': v2pi,
                     'per_sec': v2pi, 'rpm': sec2min * v2pi},
    'Angle': {'deg': deg2rad, 'rad': 1.0, 'degmin': deg2rad * sec2min, 'degsec': deg2rad * sec2hour},
    'Current': {'fA': 1e-15, 'pA': 1e-12, 'nA': 1e-9, 'uA': 1e-6, 'mA': 1e-3, 'A': 1.0, 'kA': 1e3, 'MegA': 1e6,
                'gA': 1e9, 'dBA': (dB,)},
    'Flux': {'Wb': 1.0, 'mx': 1e-8, 'vh': 3600, 'vs': 1.0},
    'Freq': {'Hz': 1.0, 'kHz': 1e3, 'MHz': 1e6, 'GHz': 1e9, 'THz': 1e12, 'rps': 1.0, 'per_sec': 1.0},
    'Inductance': {'fH': 1e-15, 'pH': 1e-12, 'nH': 1e-9, 'uH': 1e-6, 'mH': 1e-3, 'H': 1.0},
    'Length': {'fm': 1e-15, 'pm': 1e-12, 'nm': 1e-9, 'um': 1e-6, 'mm': 1e-3, 'cm': 1e-2, 'dm': 1e-1, 'meter': 1.0,
               'km': 1e3, 'uin': m2in * 1e-6, 'mil': m2in * 1e-3, 'in': m2in, 'ft': m2in * 12, 'yd': m2in * 144},
    'Mass': {'ug': 1e-9, 'mg': 1e-6, 'g': 1e-3, 'kg': 1.0, 'ton': 1000, 'oz': 0.0283495, 'lb': 0.453592 },
    'None': {'f': 1e-15, 'p': 1e-12, 'n': 1e-9, 'u': 1e-6, 'm': 1e-3, '': 1.0, 'k': 1e3, 'meg': 1e6, 'g': 1e9, 't': 1e12},
    'Resistance': {'uOhm': 1e-6, 'mOhm': 1e-3, 'ohm': 1.0, 'kOhm': 1e3, 'megohm': 1e6, 'GOhm': 1e9 },
    'Speed': {'mm_per_sec': 1e-3, 'cm_per_sec': 1e-2, 'm_per_sec': 1.0, 'km_per_sec': 1e3, 'inches_per_sec': m2in,
              'feet_per_sec': m2in * 12, 'feet_per_min': m2in * 12 * sec2min, 'km_per_min': 60e3, 'm_per_h': 3600,
              'miles_per_hour': m2miles * sec2hour, 'miles_per_minute': m2miles * sec2min, 'miles_per_sec': m2miles},
    'Time': {'fs': 1e-15, 'ps': 1e-12, 'ns': 1e-9, 'us': 1e-6, 'ms': 1e-3, 's': 1, 'min': 60, 'hour': 3600, 'day': 3600*12},
    'Torque': {'fNewtonMeter': 1e-15, 'pNewtonMeter': 1e-12, 'nNewtonMeter': 1e-9, 'uNewtonMeter': 1e-6, 'mNewtonMeter': 1e-3,
               'cNewtonMeter': 1e-2, 'NewtonMeter': 1, 'kNewtonMeter': 1e3, 'megNewtonMeter': 1e6, 'gNewtonMeter': 1e9},
    'Voltage': {'fV': 1e-15, 'pV': 1e-12, 'nV': 1e-9, 'uV': 1e-6, 'mV': 1e-3, 'V': 1.0, 'kV': 1e3, 'MegV': 1e6,
                'gV': 1e9, 'dBV': (dB,)},
    'Temperature': {'kel': 1.0, 'cel': (cel2kel,), 'fah': (fah2kel,)},
    'Power': {'fW': 1e-15, 'pW': 1e-12, 'nW': 1e-9, 'uW': 1e-6, 'mW': 1e-3, 'W': 1.0, 'kW': 1e3, 'megW': 1e6, 'gW': 1e9}
}
SI_units = {
    'AngularSpeed':  'rad_per_sec',
    'Angle': 'deg',
    'Current': 'A',
    'Flux': 'vs',
    'Freq': 'Hz',
    'Inductance': 'H',
    'Length': 'meter',
    'Mass': 'kg',
    'None': '',
    'Resistance': 'ohm',
    'Time':  's',
    'Torque': 'NewtonMeter',
    'Voltage': 'V',
    'Temperature': 'kel',
    'Power': 'W'}

unit_system_operations = {
    # Multiplication of physical domains
    'Voltage_multiply_Current': 'Power',
    'Torque_multiply_AngularSpeed': 'Power',
    'AngularSpeed_multiply_Time': 'Angle',
    'Current_multiply_Resistance': 'Voltage',
    'AngularSpeed_multiply_Inductance': 'Resistance',
    'Speed_multiply_Time': 'Length',

    # Division of Physical Domains
    'Power_divide_Voltage': 'Current',
    'Power_divide_Current': 'Voltage',
    'Power_divide_AngularSpeed': 'Torque',
    'Power_divide_Torque': 'AngularSpeed',
    'Angle_divide_AngularSpeed': 'Time',
    'Angle_divide_Time': 'AngularSpeed',
    'Voltage_divide_Current': 'Resistance',
    'Voltage_divide_Resistance': 'Current',
    'Resistance_divide_AngularSpeed': 'Inductance',
    'Resistance_divide_Inductance': 'AngularSpeed',
    'None_divide_Freq': 'Time',
    'None_divide_Time': 'Freq',
    'Length_divide_Time': 'Speed',
    'Length_divide_Speed': 'Time'
}

def _resolve_unit_system(unit_system_1, unit_system_2, operation):
    """
    Attempts to return the unit string of an arithmetic operation on Variable objects. If no resulting unit system
    is defined for a specific operation (in unit_system_operations), an empty string is returned

    Parameters
    ----------
    unit_system_1 : str
    Name of a unit system (key of AEDT_units)

    unit_system_2 : str
    Name of another unit system (key of AEDT_units)

    operation : str
    Name of an operator within the set of ["multiply" ,"divide"]

    Returns
    -------
    str

    """
    try:
        key = "{}_{}_{}".format(unit_system_1, operation, unit_system_2)
        result_unit_system = unit_system_operations[key]
        return SI_units[result_unit_system]
    except KeyError:
        return ""

class CSVDataset:

    @property
    def number_of_rows(self):
        if self._data:
            for variable, data_list in self._data.items():
                return len(data_list)
        else:
            return 0

    @property
    def number_of_columns(self):
        """ """
        return len(self._header)

    @property
    def header(self):
        """ """
        return self._header

    @property
    def data(self):
        """ """
        return self._data

    @property
    def path(self):
        """ """
        return os.path.dirname(os.path.realpath(self._csv_file))

    def __init__(self, csv_file=None, separator=None, units_dict=None, append_dict=None, valid_solutions=True, invalid_solutions=False):
        ''' Reads in the csv file and extracts header, data and units information as well as being able
            to augment the data with constant values

        :param csv_file:    Input file consisting of delimited data with first line as header file.
                            Data values support AEDT units such as e.g. 1.23Wb
        :param separator:   Data delimiter. Default is comma
        :param units_dict:  Dictionary consisting of {Variable Name: unit} used to rescale data if not in the desired
                            unit system
        :param append_dict: Dictionary {New Variable Name: value} Additional variable with constant value for all
                            Data points. (Used to add multiple sweeps to one result file
        '''
        self._header = []
        self._data = {}
        self._unit_dict = {}
        self._append_dict = {}

        # Set the index counter explicitly to zero
        self._index = 0

        if separator:
            self._separator = separator
        else:
            self._separator = ","

        if units_dict:
            self._unit_dict = units_dict

        if append_dict:
            self._append_dict = append_dict

        self._csv_file = csv_file
        if csv_file:
            with open(csv_file, 'r') as fi:
                file_data = fi.readlines()
                for line in file_data:
                    if self._header:
                        line_data = line.strip().split(self._separator)
                        # Check for invalid data in the line (fields with 'nan')
                        if not 'nan' in line_data:
                            for j, value in enumerate(line_data):
                                var_name = self._header[j]
                                if var_name in self._unit_dict:
                                    var_value = Variable(value).rescale_to(self._unit_dict[var_name]).numeric_value
                                else:
                                    var_value = Variable(value).value
                                self._data[var_name].append(var_value)

                            # Add augmented quantities
                            for entry in self._append_dict:
                                var_value_str = self._append_dict[entry]
                                numeric_value = Variable(var_value_str).numeric_value
                                self._data[entry].append(numeric_value)

                    else:
                        self._header = line.strip().split(",")
                        for additional_quantity_name in self._append_dict:
                            self._header.append(additional_quantity_name)
                        for quantity_name in self._header:
                            self._data[quantity_name] = []

        pass

    @aedt_exception_handler
    def __getitem__(self, item):
        variable_list = item.split(',')
        data_out = CSVDataset()
        for variable in variable_list:
            found_variable = False
            for key_string in self._data:
                if variable in key_string:
                    found_variable = True
                    break
            assert found_variable, "Input string {} is not a key of the data dictionary!".format(variable)
            data_out._data[variable] = self._data[key_string]
            data_out._header.append(variable)
        return data_out

    @aedt_exception_handler
    def __add__(self, other):
        assert self.number_of_columns == other.number_of_columns, "Inconsistent number of columns"
        # Create a new object to return, avoiding changing the original inputs
        new_dataset = CSVDataset()
        # Add empty columns to new_dataset
        for column in self._data:
            new_dataset._data[column] = []

        # Add the data from 'self' to a the new dataset
        for column, row_data in self.data.items():
            for value in row_data:
                new_dataset._data[column].append(value)

        # Add the data from 'other' to a the new dataset
        for column, row_data in other.data.items():
            for value in row_data:
                new_dataset._data[column].append(value)

        return new_dataset

    def __iadd__(self, other):
        ''' Incrementally add one CSVDataset to another. This assumes that the number of columns are the same,
            or that self is an empty CSVDataset. No checking for equivalency of units/variable names is done'''

        # Handle the case of an empty data set and create empty lists for the column data
        if self.number_of_columns == 0:
            self._header = other.header
            for column in other.data:
                self._data[column] = []

        assert self.number_of_columns == other.number_of_columns, "Inconsistent number of columns"

        # Append the data from 'other'
        for column, row_data in other.data.items():
            for value in row_data:
                self._data[column].append(value)


        return self

    # Called when iteration is initialized
    def __iter__(self):

        self._index = 0
        return self

    # Create an iterator to yield the row data as a string as we loop through the object
    def __next__(self):
        if self._index < (self.number_of_rows-1):
            output = []
            for column in self._header:
                string_value = str(self._data[column][self._index])
                output.append(string_value)
            output_string = " ".join(output)
            self._index += 1
        else:
            raise StopIteration

        return output_string

    def next(self):
        """ """
        return self.__next__()

@aedt_exception_handler
def decompose_variable_value(variable_value):
    """

    Parameters
    ----------
    variable_value :

    Returns
    -------

    """
    # set default return values - then check for valid units
    float_value = variable_value
    units =''

    if isinstance(variable_value, numbers.Number):
        float_value = variable_value
    elif isinstance(variable_value, str) and variable_value != 'nan':
        try:
            # Handle a numerical value in string form
            float_value = float(variable_value)
        except ValueError:
            # search for a valid units string at the end of the variable_value
            loc = re.search('[a-z_A-Z]+$', variable_value)
            if loc:
                loc_units = loc.span()[0]
                extract_units = variable_value[loc_units:]
                if unit_system(extract_units):
                    try:
                        float_value = float(variable_value[0:loc_units])
                        units = extract_units
                    except ValueError:
                        float_value = variable_value

    return float_value, units


class VariableManager(object):
    """
    Variable Manager Class

    This object manages the design properties (local variables in a design) and project variables (Variables defined
    at the project level (starting with $). This class provides access to all or a subset of the variables. Manipulation
    of the numerical or string definition of the variable values is provided in the 'Variable' class.

    Attributes
    ------
    variables : dict of Variable
        Dictionary of all design properties and project variables in the active design

    design_variables : dict of Variable
        Dictionary of all design properties (local variables) in the active design

    project_variables : dict of Variable
        Dictionary of all project variables (starting with '$') available to the active design (key by variable name)

    dependent_variables : dict of Variable
       Dictionary of all dependent variables available to the active design (key by variable name)

    independent_variables : dict of Variable
       Dictionary of all independent variables (constant numeric values) available to the active design (key by variable name)

    independent_design_variables : dict of Variable

    independent_project_variables : dict of Variable

    variable_names : list of str
    project_variable_names : list of str
    design_variable_names : list of str
    dependent_variable_names : list of str
        List of all dependent variable names within the project
            independent_variable_names : list of str
        List of all independent variable names within the project (can be sweep variables for optimetrics)
    independent_project_variable_names : list of str
        List of all independent variable names within the project (can be sweep variables for optimetrics)
    independent_design_variable_names : list of str
        List of all independent variable names within the project (can be sweep variables for optimetrics)


    Examples
    ------

    >>> from pyaedt.maxwell import Maxwell3d
    >>> from pyaedt.desktop import Desktop
    >>> d = Desktop()
    >>> aedtapp = Maxwell3d()
    >>> # Define some test variables
    >>> aedtapp["Var1"] = 3
    >>> aedtapp["Var2"] = "12deg"
    >>> aedtapp["Var3"] = "Var1 * Var2"
    >>> aedtapp["$PrjVar1"] = "pi"

    >>> # Get the variable manager for the active design
    >>> v = aedtapp.variable_manager
    >>> # get a dictionary of all project & design variables
    >>> v.variables

    {'Var1': <pyaedt.application.Variables.Variable at 0x2661f34c448>,
     'Var2': <pyaedt.application.Variables.Variable at 0x2661f34c308>,
     'Var3': <pyaedt.application.Variables.Expression at 0x2661f34cb48>,
     '$PrjVar1': <pyaedt.application.Variables.Expression at 0x2661f34cc48>}

    >>> # get a dictionary of just the design variables
    >>> v.design_variables

    {'Var1': <pyaedt.application.Variables.Variable at 0x2661f339508>,
     'Var2': <pyaedt.application.Variables.Variable at 0x2661f3415c8>,
     'Var3': <pyaedt.application.Variables.Expression at 0x2661f341808>}

    >>> # get a dictionary of just the independent design variables
    >>> v.independent_design_variables
    {'Var1': <pyaedt.application.Variables.Variable at 0x2661f335d08>,
     'Var2': <pyaedt.application.Variables.Variable at 0x2661f3557c8>}

    See Also
    --------
    Variable class
    """
    @property
    def variables(self):
        """
        Returns a dictionary of Variable objects for each project variable and each
        design property in the active design

        Returns
        -------
            dict of Variable
        """
        return self._variable_dict([self.odesign, self.oproject])

    @property
    def design_variables(self):
        return self._variable_dict([self.odesign])

    @property
    def project_variables(self):
        return self._variable_dict([self.oproject])

    @property
    def independent_variables(self):
        return self._variable_dict([self.odesign, self.oproject], dependent=False)

    @property
    def independent_project_variables(self):
        return self._variable_dict([self.oproject], dependent=False)

    @property
    def independent_design_variables(self):
        return self._variable_dict([self.odesign], dependent=False)

    @property
    def dependent_variables(self):
        return self._variable_dict([self.odesign, self.oproject], independent=False)

    @property
    def variable_names(self):
        return [var_name for var_name in self.variables]

    @property
    def project_variable_names(self):
        return [var_name for var_name in self.project_variables]

    @property
    def independent_project_variable_names(self):
        return [var_name for var_name in self.independent_project_variables]

    @property
    def design_variable_names(self):
        return [var_name for var_name in self.design_variables]

    @property
    def independent_design_variable_names(self):
        return [var_name for var_name in self.independent_design_variables]

    @property
    def independent_variable_names(self):
        return [var_name for var_name in self.independent_variables]

    @property
    def dependent_variable_names(self):
        return [var_name for var_name in self.dependent_variables]

    @property
    def oproject(self):
        return self._parent._oproject

    @property
    def odesign(self):
        return self._parent._odesign

    @property
    def messenger(self):
        return self._parent._messenger

    def __init__(self, parent):

        # Global Desktop Environment
        self._parent = parent

    @aedt_exception_handler
    def __getitem__(self, variable_name):
        return self.variables[variable_name]

    @aedt_exception_handler
    def __setitem__(self, variable, value):
        self.set_variable(variable, value)
        return True

    @aedt_exception_handler
    def _variable_dict(self, object_list, dependent=True, independent=True):
        """

        Parameters
        ----------
        object_list :

        dependent :
             (Default value = True)
        independent :
             (Default value = True)

        Returns
        -------

        """
        var_dict = {}
        for obj in object_list:
            for variable_name in obj.GetVariables():
                variable_expression = self.get_expression(variable_name)
                try:
                    value = Variable(variable_expression)
                    if independent and isinstance(value.value, numbers.Number) :
                        var_dict[variable_name] = value
                    elif dependent and type(value.value) is str:
                        float_value = self._parent.get_evaluated_value(variable_name)
                        var_dict[variable_name] = Expression(variable_expression, float_value)
                except:
                    if dependent:
                        float_value = self._parent.get_evaluated_value(variable_name)
                        var_dict[variable_name] = Expression(variable_expression, float_value)

        return var_dict

    @aedt_exception_handler
    def get_expression(self, variable_name):
        """ Return the variable value of a project or design variable as a string"""
        return self.aedt_object(variable_name).GetVariableValue(variable_name)

    @aedt_exception_handler
    def aedt_object(self, variable):
        if variable[0] == "$":
            return self.oproject
        else:
            return self.odesign

    @aedt_exception_handler
    def set_variable(self, variable_name, expression=None, readonly=False, hidden=False,
                     description=None, overwrite=True):
        """Set the value of a project variable or design property

        If the specified variable name does not exist, create a new property and set the value.

        Parameters
        ----------
        variable_name : str
        Name of a design property or project variable ($var) to be set.

        expression : str
        Valid string expression within the AEDT design/project structure, e.g. "3*cos(34deg)"

        readonly : bool, default=False
        Set the "read-only" flag for the property

        hidden : bool, default=False
        Set the "hidden" flag for the property

        description : str, default=None
        Description string to be displayed in the properties window

        overwrite : bool, default=True
        Overwrite the value of an existing property, if False - ignore this command

        Returns
        -------
        bool

        Examples
        --------
        Set the value of the design property p1 to "10mm" - create the property if it does not exist already
        >>> aedtapp.variable_manager.set_variable("p1", expression="10mm")

        Set the value of the design property p1 to "20mm" only if the property does not exist already
        >>> aedtapp.variable_manager.set_variable("p1", expression="20mm", overwrite=False)

        Set the value of the design property p2 to "10mm"  - create the property if it does not exist already and
        add additional flags "read-only" and "hidden" plus a description text
        >>> aedtapp.variable_manager.set_variable(variable_name="p2", expression="10mm", readonly=True, hidden=True,
        ...                                       description="This is a description of this variable")

        Set the value of the project variable $p1 to "30mm" - create the variable if it does not exist
        >>> aedtapp.variable_manager.set_variable["$p1"] == "30mm"

        """
        if not description:
            description = ""

        desktop_object = self.aedt_object(variable_name)
        test = desktop_object.GetName()
        proj_name = self.oproject.GetName()
        var_type = "Project" if test == proj_name else "Local"

        prop_type="VariableProp"

        if isinstance(expression, str):
            # Handle string type variable (including arbitrary expression)# Handle input type variable
            variable = expression
        elif isinstance(expression, Variable):
            # Handle input type variable
            variable = expression.string_value
        elif isinstance(expression, numbers.Number):
            # Handle input type int/float, etc (including numeric 0)
            variable = str(expression)
        # Handle None, "" as Separator
        elif isinstance(expression, list):
            variable = str(expression)
        elif not expression:
            prop_type = "SeparatorProp"
            variable = ""
            try:
                if self.delete_separator(variable_name):
                    desktop_object.Undo()
                    self.messenger.clear_messages()
                    return
            except:
                pass
        else:
            raise Exception("Unhandled input type to design/project variable.")

        if "post" in variable_name.lower()[0:5]:
            prop_type = "PostProcessingVariableProp"

        # Get all design and project variables in lower case for a case-sensitive comparison
        var_list = desktop_object.GetVariables()
        lower_case_vars = [var_name.lower() for var_name in var_list]

        if variable_name.lower() not in lower_case_vars:

            desktop_object.ChangeProperty(
                [
                    "NAME:AllTabs",
                    [
                        "NAME:{0}VariableTab".format(var_type),
                        [
                            "NAME:PropServers",
                            "{0}Variables".format(var_type)
                        ],
                        [
                            "NAME:NewProps",
                            [
                                "NAME:" + variable_name,
                                "PropType:=", prop_type,
                                "UserDef:=", True,
                                "Value:=", variable,
                                "Description:=", description,
                                "ReadOnly:=", readonly,
                                "Hidden:=", hidden
                            ]
                        ]
                    ]
                ])
        elif overwrite:
            desktop_object.ChangeProperty(
                [
                    "NAME:AllTabs",
                    [
                        "NAME:{}VariableTab".format(var_type),
                        [
                            "NAME:PropServers",
                            "{}Variables".format(var_type)
                        ],
                        [
                            "NAME:ChangedProps",
                            [
                                "NAME:" + variable_name,
                                "Value:="		, variable,
                                "Description:="		, description,
                                "ReadOnly:="		, readonly,
                                "Hidden:="		, hidden
                            ]
                        ]
                    ]
                ])

        return True

    @aedt_exception_handler
    def delete_separator(self, separator_name):
        """
            Deletes a separator from either the active project or design

        :param separator_name: Separator text
        :return: True if the separator exists and can be deleted, False otherwise
        """
        object_list = [ (self.odesign, "Local"),
                        (self.oproject, "Project")]

        for object_tuple in object_list:
            desktop_object = object_tuple[0]
            var_type = object_tuple[1]
            try:
                desktop_object.ChangeProperty(
                    [
                        "NAME:AllTabs",
                        [
                            "NAME:{0}VariableTab".format(var_type),
                            [
                                "NAME:PropServers",
                                "{0}Variables".format(var_type)
                            ],
                            [
                                "NAME:DeletedProps",
                                separator_name
                            ]
                        ]
                    ])
                return True
            except:
                pass
        return False

    @aedt_exception_handler
    def delete_variable(self, sVarName):
        """

        Parameters
        ----------
        sVarName :


        Returns
        -------

        """
        desktop_object = self.aedt_object(sVarName)
        var_type = "Project" if desktop_object == self.oproject else "Local"

        var_list = desktop_object.GetVariables()
        lower_case_vars = [var_name.lower() for var_name in var_list]

        if sVarName.lower() in lower_case_vars:
            try:
                desktop_object.ChangeProperty(
                    [
                        "NAME:AllTabs",
                        [
                            "NAME:{0}VariableTab".format(var_type),
                            [
                                "NAME:PropServers",
                                "{0}Variables".format(var_type)
                            ],
                            [
                                "NAME:DeletedProps",
                                sVarName
                            ]
                        ]
                    ])
                return True
            except:
                pass
        return False


class Variable(object):
    """
    Variable Class. It stores Variables and allows operations on it

    This object handles the contents of design properties and project variables.

    Parameters
    ----------
    value
    units

    Attributes
    ------
    units : str
        The defined variable units
    value : float
        The numerical value of the variable in SI-Units

    Examples
    --------
    >>> from pyaedt.application.Variables import Variable

    Define a variable using a string-value consistent with the AEDT properties
    >>> v = Variable("45mm")

    A unitless variable with value 3.0
    >>> v = Variable(3.0)

    A variable defined by a numeric result and a unit string
    >>> v = Variable(3.0 * 4.5, units="mm")
    >>> assert v.numeric_value = 13.5
    >>> assert v.units = "mm"

    """

    def __init__(self, value, units=None):

        if units:
            if unit_system(units):
                specified_units = units

        self._units = None
        self._expression = value
        self._value, self._units = decompose_variable_value(value)

        # If units have been specified, check for a conflict and otherwise use the specified unit system
        if units:
            assert not self._units, \
                "Inconsistent unit specification {} made vs identified units {}".format(specified_units, self._units)
            self._units = specified_units

        if isinstance(self._value, numbers.Number):
            scale = AEDT_units[self.unit_system][self._units]
            if isinstance(scale, tuple):
                self._value = scale[0](self._value, inverse=False)
            else:
                self._value =  self._value * scale

    @property
    def unit_system(self):
        """ Return the unit system of the expression as a string """
        return unit_system(self._units)

    @property
    def units(self):
        return self._units

    @property
    def value(self):
        return self._value

    @property
    def numeric_value(self):
        """ Return the numeric part of the expression as a float value """
        if isinstance(self._value, numbers.Number):
            scale = AEDT_units[self.unit_system][self._units]
        if isinstance(scale, tuple):
            return scale[0](self._value, True)
        else:
            return self._value / scale

    @property
    def string_value(self):
        """
        Concatenate the numeric value with the unit and return as string. There can be differences in the
        numeric display, e.g. 10mm in the modeler could be 10.0mm returned from string_value
        """
        return ('{}{}').format(self.numeric_value, self._units)

    @aedt_exception_handler
    def rescale_to(self, units):
        """ Change the unit system of the expression to a new unit within the same unit system

        Examples
        --------
        >>> from pyaedt.application.Variables import Variable

        >>> v = Variable("10W")
        >>> assert v.numeric_value == 10
        >>> assert v.units == "W"
        >>> v.rescale_to("kW")
        >>> assert v.numeric_value == 0.01
        >>> assert v.units == "kW"

        """
        new_unit_system = unit_system(units)
        assert new_unit_system == self.unit_system, \
            "New unit system {0} inconsistent with current unit system {1}."
        self._units = units
        return self


    @aedt_exception_handler
    def format(self, format):
        """Return the string value with specified numerical formatting

        Parameters
        ----------
        format : str
            Format string for the numeric value of the string e.g. '06.2f' (see https://pyformat.info/)

        Returns
        -------
        str

        Examples
        --------
        >>> from pyaedt.application.Variables import Variable

        >>> v = Variable("10W")
        >>> assert v.format("f") == '10.000000W'
        >>> assert v.format("06.2f") == '010.00W'
        >>> assert v.format("6.2f") == ' 10.00W'

        """
        return ('{0:' + format + '}{1}').format(self.numeric_value, self._units)

    @aedt_exception_handler
    def __mul__(self, other):
        """Multiply the Variable with a number or another Variable and return a new object

        Parameters
        ---------
        other : numbers.Number or Variable
            object to be multiplied

        Returns
        -------
        Variable

        Examples
        --------
        >>> from pyaedt.application.Variables import Variable

        Multiply 'Length' by unitless ('None') to obtain 'Length'. A numerical value is also considered to be unitless
        >>> v1 = Variable("10mm")
        >>> v2 = Variable(3)
        >>> result_1 = v1 * v2
        >>> result_2 = v1 * 3
        >>> assert result_1.numeric_value == 30.0
        >>> assert result_1.unit_system == "Length"
        >>> assert result_2.numeric_value == result_1.numeric_value
        >>> assert result_2.unit_system == result_1.unit_system

        Multiply 'Voltage' times 'Current' to obtain 'Power'
        >>> v3 = Variable("3mA")
        >>> v4 = Variable("40V")
        >>> result_3 = v3 * v4
        >>> assert result_3.numeric_value == 0.12
        >>> assert result_3.units == "W"
        >>> assert result_3.unit_system == "Power"

        """
        assert isinstance(other, numbers.Number) or isinstance(other, Variable), "Multiplier must be a scalar quantity or a variable!"
        if isinstance(other, numbers.Number):
            result_value = self.numeric_value * other
            result_units = self.units
        else:
            if self.unit_system == "None":
                return self.numeric_value * other
            elif other.unit_system == "None":
                return other.numeric_value * self
            else:
                result_value = self.value * other.value
                result_units = _resolve_unit_system(self.unit_system, other.unit_system, "multiply")
                if not result_units:
                    result_units = _resolve_unit_system(other.unit_system, self.unit_system, "multiply")

        return Variable("{}{}".format(result_value, result_units))

    __rmul__ = __mul__

    @aedt_exception_handler
    def __add__(self, other):
        """Add the Variable with another Variable and return a new object

        Parameters
        ---------
        other : Variable
            object to be multiplied

        Returns
        -------
        Variable

        Examples
        --------
        >>> from pyaedt.application.Variables import Variable

        >>> v1 = Variable("3mA")
        >>> v2 = Variable("10A")
        >>> result = v1 + v2
        >>> assert result.numeric_value == 10.003
        >>> assert result.units == "A"
        >>> assert result.unit_system == "Current"

        """
        assert isinstance(other, Variable), "Can only add Variable with Variable"
        assert self.unit_system == other.unit_system, "Only Variable objects with the same unit system can be added"
        result_value = self.value + other.value
        result_units = SI_units[self.unit_system]
        # If the units of the two operands are different, return SI-Units
        result_variable = Variable("{}{}".format(result_value, result_units))

        # If the units of both operands are the same, return those units
        if self.units == other.units:
            result_variable.rescale_to(self.units)

        return result_variable

    @aedt_exception_handler
    def __sub__(self, other):
        """Subtract another Variable from the Variable and return a new object

        Parameters
        ---------
        other : Variable
            object to be subtracted

        Returns
        -------
        Variable

        Examples
        --------
        >>> from pyaedt.application.Variables import Variable
        >>> v3 = Variable("3mA")
        >>> v4 = Variable("10A")
        >>> result_2 = v3 - v4
        >>> assert result_2.numeric_value == -9.997
        >>> assert result_2.units == "A"
        >>> assert result_2.unit_system == "Current"

        """
        assert isinstance(other, Variable), "Can only subtract Variable from Variable"
        assert self.unit_system == other.unit_system, "Only Variable objects with the same unit system can be added"
        result_value = self.value - other.value
        result_units = SI_units[self.unit_system]
        # If the units of the two operands are different, return SI-Units
        result_variable = Variable("{}{}".format(result_value, result_units))

        # If the units of both operands are the same, return those units
        if self.units == other.units:
            result_variable.rescale_to(self.units)

        return result_variable

    # Python 3.x version
    @aedt_exception_handler
    def __truediv__(self, other):
        """Divide the Variable with a number or another Variable and return a new object

        Parameters
        ---------
        other : numbers.Number or Variable
            divisor object

        Returns
        -------
        Variable

        Examples
        --------
        Divide a variable with units 'W' with a variable with units 'V'. Automatically resolve the new units to 'A'
        >>> from pyaedt.application.Variables import Variable
        >>> v1 = Variable("10W")
        >>> v2 = Variable("40V")
        >>> result = v1 / v2
        >>> assert result_1.numeric_value == 0.25
        >>> assert result_1.units == "A"
        >>> assert result_1.unit_system == "Current"

        """
        assert isinstance(other, numbers.Number) or isinstance(other, Variable), "Divisor must be a scalar quantity or a variable!"
        if isinstance(other, numbers.Number):
            result_value = self.numeric_value / other
            result_units = self.units
        else:
            result_value = self.value / other.value
            result_units = _resolve_unit_system(self.unit_system, other.unit_system, "divide")

        return Variable("{}{}".format(result_value, result_units))

    # Python 2.7 version
    @aedt_exception_handler
    def __div__(self, other):
        return self.__truediv__(other)

    @aedt_exception_handler
    def __rtruediv__(self, other):
        """Divide the other object by this object

        Parameters
        ---------
        other : numbers.Number or Variable
            Dividend object

        Returns
        -------
        Variable

        Examples
        --------
        Divide a number by a Variable with units 's'. Automatically determine the result is in 'Hz'
        >>> from pyaedt.application.Core.Variables import Variable
        >>> v = Variable("1s")
        >>> result = 3.0 / v
        >>> assert result.numeric_value == 3.0
        >>> assert result.units == "Hz"
        >>> assert result.unit_system == "Freq"

        """
        assert isinstance(other, numbers.Number), "Dividend must be a numerical quantity!"
        result_value = other / self.value
        result_units = _resolve_unit_system("None", self.unit_system, "divide")
        return Variable("{}{}".format(result_value, result_units))

    # Python 2.7 version
    @aedt_exception_handler
    def __div__(self, other):
        return self.__rtruediv__(other)

class Expression(Variable, object):
    """Class to manipulate variable expressions"""

    def __init__(self, expression, float_value):
        self._expression = expression
        self._value = float_value
        # TODO. Try to identify the unit system from the expression !
        try:
            value, units = decompose_variable_value(expression)
            self._units = units
        except:
            self._units = ''

    @property
    def expression(self):
        """ """
        return str(self._expression)


class DataSet(object):
    """Class for Dataset Management"""
    def __init__(self, parent, name, x, y, z=None, v=None, xunit='', yunit='', zunit='', vunit=''):
        self._parent = parent
        self.name = name
        self.x = x
        self.y = y
        self.z = z
        self.v = v
        self.xunit = xunit
        self.yunit = yunit
        self.zunit = zunit
        self.vunit = vunit

    @aedt_exception_handler
    def _args(self):
        """ """
        arg = []
        arg.append("Name:" + self.name)
        arg2 = ["Name:Coordinates"]
        if self.z is None:
            arg2.append(["NAME:DimUnits", self.xunit, self.yunit])
        elif self.v is not None and self.name[0] == "$":
            arg2.append(["NAME:DimUnits", self.xunit, self.yunit, self.zunit, self.vunit])
        else:
            return False
        if self.z and self.name[0] == "$":
            x, y, z, v = (list(t) for t in zip(*sorted(zip(self.x, self.y, self.z, self.v))))
        else:
            x, y = (list(t) for t in zip(*sorted(zip(self.x, self.y))))
        for i in range(len(x)):
            arg3 = []
            arg3.append("NAME:Coordinate")
            arg4 = ["NAME:CoordPoint"]
            arg4.append(float(x[i]))
            arg4.append(float(y[i]))
            if self.z and self.name[0] == "$":
                arg4.append(float(z[i]))
                arg4.append(float(v[i]))
            arg3.append(arg4)
            arg2.append(arg3)
        arg.append(arg2)
        return arg

    @aedt_exception_handler
    def create(self):
        """Create a new Dataset


        Parameters
        ----------

        Returns
        -------

        """
        if self.name[0] == "$":
            self._parent._oproject.AddDataset(self._args())
        else:
            self._parent.odesign.AddDataset(self._args())
        return True

    @aedt_exception_handler
    def add_point(self, x, y, z=None, v=None):
        """Add a new point to the existing dataset

        Parameters
        ----------
        x :
            float
        y :
            float
        z :
            float or None (Default value = None)
        v :
            float or None (Default value = None)

        Returns
        -------
        type
            Bool

        """
        self.x.append(x)
        self.y.append(y)
        if self.z and self.v:
            self.z.append(z)
            self.v.append(v)

        return self.update()

    @aedt_exception_handler
    def remove_point_from_x(self, x):
        """Remove Point from given x value

        Parameters
        ----------
        x :
            float

        Returns
        -------
        type
            Bool

        """
        if x not in self.x:
            self._parent._messenger.add_error_message("Value {} not found.".format(x))
            return False
        id_to_remove = self.x.index(x)
        return self.remove_point_from_index(id_to_remove)


    @aedt_exception_handler
    def remove_point_from_index(self, id_to_remove):
        """Remove Point from a specific index.

        Parameters
        ----------
        id_to_remove :
            index to remove (integer)

        Returns
        -------
        type
            Bool

        """
        if id_to_remove < len(self.x) > 2:
            self.x.pop(id_to_remove)
            self.y.pop(id_to_remove)
            if self.z and self.v:
                self.z.pop(id_to_remove)
                self.v.pop(id_to_remove)
            return self.update()
        self._parent._messenger.add_error_message("cannot Remove {} index.".format(id_to_remove))
        return False

    @aedt_exception_handler
    def update(self):
        """Update Dataset"""
        args = self._args()
        if not args:
            return False
        if self.name[0] == "$":
            self._parent._oproject.EditDataset(self.name, self._args())
        else:
            self._parent.odesign.EditDataset(self.name, self._args())
        return True

    @aedt_exception_handler
    def delete(self):
        """Delete the dataset."""
        if self.name[0] == "$":
            self._parent._oproject.DeleteDataset(self.name)
            del self._parent.project_datasets[self.name]
        else:
            self._parent.odesign.DeleteDataset(self.name)
            del self._parent.project_datasets[self.name]
        return True

    @aedt_exception_handler
    def export(self, dataset_path=None):
        """Export dataset

        Parameters
        ----------
        dataset_path : str, optional
            Path where to export.  If ``None`` the dataset will be
            exported to project_path (Default value = None)

        """
        if not dataset_path:
            dataset_path = os.path.join(self._parent.project_path, self.name +".tab")
        if self.name[0] == "$":
            self._parent._oproject.ExportDataset(self.name, dataset_path)
        else:
            self._parent.odesign.ExportDataset(self.name, dataset_path)
        return True




"""This module contains these classes: ``Hfss3dLayout`` and ``SweepString``."""

from __future__ import absolute_import
import os

from .application.Analysis3DLayout import FieldAnalysis3DLayout
from .desktop import exception_to_desktop
from .modules.SolveSetup import Setup3DLayout
from .generic.general_methods import generate_unique_name, aedt_exception_handler



class SweepString(object):
    """Generate a sweep string like this example: ``"LIN 10GHz 20GHz 0.05GHz LINC 20GHz 30GHz 10 DEC 30GHz 40GHz 10 40GHz``"""
    def __init__(self, unit='GHz'):
        """

        :param unit:
        """
        self.unit = unit
        self.sweep_string = ""

    @aedt_exception_handler
    def add_sweep(self, sweep, line_type, unit=None):
        """Add a sweep line to the string.

        Parameters
        ----------
        sweep : list
            List of frequencies,
            if linear_step [start, stop, step]
            if linear_count [start, stop, number of steps]
            if log_scale [start, stop, samples]
            if single [f1, f2,... fn]
        line_type :
            linear_step", "linear_count", "log_scale", "single"
        unit : str
            Units such as ``"MHz"``, ``"GHz"``, and so on. The default is ``None``.

        Returns
        -------

        """
        if not unit:
            unit = self.unit
        if self.sweep_string != "":
            self.sweep_string += " "
        if line_type == "linear_step" and len(sweep) == 3:
            string = "LIN " + str(sweep[0]) + unit + " " + str(sweep[1]) + unit + " " + str(sweep[2]) + unit
        elif line_type == "linear_count" and len(sweep) == 3:
            string = "LINC " + str(sweep[0]) + unit + " " + str(sweep[1]) + unit + " " + str(sweep[2])
        elif line_type == "log_scale" and len(sweep) == 3:
            string = "DEC " + str(sweep[0]) + unit + " " + str(sweep[1]) + unit + " " + str(sweep[2])
        elif line_type == "single":
            string = ""
            for f in sweep:
                string += str(f) + unit + " "
            string = string[:-1]
        else:
            raise ValueError('Wrong format for sweep list!')
        self.sweep_string += string

    @aedt_exception_handler
    def get_string(self):
        """ """
        if self.sweep_string:
            return self.sweep_string
        else:
            return None


class Hfss3dLayout(FieldAnalysis3DLayout):
    """HFSS 3D Layout instance interface.

    This class contains all HFSS 3D Layout functionalities. It
    inherits all objects that belong to HFSS 3D Layout, including EDB
    API queries.

    Parameters
    ----------
    projectname : str
        Name of the project to select or the full path to the project
        or AEDTZ archive to open.  If ``None``, try to get the active
        project and, if none exists, create an empty project.
    designname : str
        Name of the design to select. If ``None``, try to get the
        active design and, if none exists, create an empty design.
    solution_type : str
        Solution type to apply to the design. If ``None``, use the
        default.
    setup_name :
        Name of the setup to use as the nominal. If ``None``, the
        active setup is used or nothing is used.

    Examples
    --------
    Create an ``Hfss3dLayout`` object and connect to an existing HFSS
    design or create a new HFSS design if one does not exist.

    >>> from pyaedt import Hfss3dLayout
    >>> aedtapp = Hfss3dLayout()

    Create an ``Hfss3dLayout`` object and link to a project named
    ``projectname``. If this project does not exist, create one with
    this name.

    >>> aedtapp = Hfss3dLayout(projectname)

    Create an ``Hfss3dLayout`` object and link to a design named
    ``designname`` in a project named ``projectname``.

    >>> aedtapp = Hfss3dLayout(projectname,designame)

    Create an ``Hfss3dLayout`` object and open the specified project.

    >>> aedtapp = Hfss3dLayout("myfile.aedt")

    Create a ``Desktop on 2021R1`` object and then create a
    ``Hfss3dLayout`` object and open the specified project.

    >>> aedtapp = Hfss3dLayout(specified_version="2021.1", projectname="myfile.aedt")
    """

    def __init__(self, projectname=None, designname=None, solution_type=None, setup_name=None,
                 specified_version=None, NG=False, AlwaysNew=False, release_on_exit=False):
        FieldAnalysis3DLayout.__init__(self, "HFSS 3D Layout Design", projectname, designname, solution_type,
                                       setup_name, specified_version, NG, AlwaysNew, release_on_exit)

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        ''' Push exit up to parent object Design '''
        if ex_type:
            exception_to_desktop(self, ex_value, ex_traceback)

    @aedt_exception_handler
    def create_edge_port(self, primivitivename, edgenumber, iscircuit=True):
        """Create a new edge port.

        Parameters
        ----------
        primivitivename : str
            Name of the primitive.
        edgenumber :
            Edge number on which to create the port.
        iscircuit : bool
            Indicates if it is a circuit port. The default is ``False``.

        Returns
        -------
        type
            Port name

        """
        listp = self.port_list
        self.modeler.oeditor.CreateEdgePort(
            ["NAME:Contents", "edge:=", ["et:=", "pe", "prim:=", primivitivename, "edge:=", edgenumber], "circuit:=",
             iscircuit, "btype:=", 0])
        listnew = self.port_list
        a=[i for i in listnew if i not in listp]
        if len(a)>0:
            return a[0]
        else:
            return False

    @aedt_exception_handler
    def create_coax_port(self, vianame, layer, xstart, xend,ystart, yend, archeight=0, arcrad=0, isexternal=True):
        """Create a new coax port.

        Parameters
        ----------
        vianame : str
            Name of the via on which to create the port.
        layer : str
            Name of the layer.
        xstart :
            X-axis position of the pin.
        ystart :
            Y-axis position of the pin.
        xend :
            X-axis position of the pin.
        yend :
            Y-axis end position of the pin.
        archeight :
            Arc height. The default is ``0``.
        arcrad :
            Rotation of the pin in rad. The default is ``0``.
        isexternal : bool
            Indicates if the pin is external. If ''True,`` the pin is external. If ''False,`` the pin is internal. The default is ``True``.

        Returns
        -------
        type
            Name of the port | False error

        """
        listp = self.port_list
        if type(layer) is str:
            layerid = self.modeler.layers.layer_id(layer)
        else:
            layerid = layer
        self.modeler.oeditor.CreateEdgePort(["NAME:Contents", "edge:=",
                                             ["et:=", "pse", "sel:=", vianame, "layer:=", layerid, "sx:=", xstart, "sy:=",
                                              ystart, "ex:=", xend, "ey:=", yend, "h:=", archeight, "rad:=", arcrad],
                                             "external:=", isexternal])
        listnew = self.port_list
        a=[i for i in listnew if i not in listp]
        if len(a)>0:
            return a[0]
        else:
            return False

    @aedt_exception_handler
    def create_pin_port(self,name,xpos=0, ypos=0, rotation=0, top_layer=None, bot_layer=None):
        """Create a new pin port.

        Parameters
        ----------
        name : str
            Name of the pin port.
        xpos :
            X-axis position of the pin. The default is ``0``.
        ypos :
            Y-axis position of the pin. The default is ``0``.
        rotation :
            Rotation of the pin in degrees. The default is ``0``.
        top_layer : str
            Top layer of the pin. If ``None``, it is automatically assigned to the top. The default is ``None``.
        bot_layer : str
            Bottom layer of the pin. If ``None``, it is automatically assigned to the bottom. The default is ``None``.

        Returns
        -------
        type
            True

        """
        self.modeler.layers.refresh_all_layers()
        layers = self.modeler.layers.all_signal_layers
        if not top_layer:
            top_layer = layers[0]
        if not bot_layer:
            bot_layer = layers[len(layers)-1]
        self.modeler.oeditor.CreatePin(
            [
                "NAME:Contents",
                [
                    "NAME:Port",
                    "Name:="	, name
                ],
                "ReferencedPadstack:="	, "Padstacks:NoPad SMT East",
                "vposition:=",
                ["x:=", str(xpos) + self.modeler.model_units, "y:=", str(ypos) + self.modeler.model_units],
                "vrotation:="		, [str(rotation) + "deg"],
                "overrides hole:="	, False,
                "hole diameter:="	, ["0mm"],
                "Pin:="			, True,
                "highest_layer:="	, top_layer,
                "lowest_layer:="	, bot_layer
            ])
        return True

    @aedt_exception_handler
    def delete_port(self, portname):
        """

        Parameters
        ----------
        portname : str
        The name of the port.          

        Returns
        -------

        """
        self.oexcitation.Delete(portname)
        return True

    @aedt_exception_handler
    def import_edb(self, edb_full_path):
        """

        Parameters
        ----------
        edb_full_path : str
            

        Returns
        -------

        """
        oTool = self.odesktop.GetTool("ImportExport")
        oTool.ImportEDB(edb_full_path)
        self.oproject = self.odesktop.GetActiveProject().GetName()
        self.odesign = self.odesktop.GetActiveProject().GetActiveDesign().GetName().split(";")[1]
        return True

    @aedt_exception_handler
    def validate_full_design(self, name=None, outputdir=None, ports=None):
        """Validate the design based on the expected value and save the information in the log file.

        Parameters
        ----------
        name : str
            Name of the design to validate. The default is ``None``.
        outputdir : str
            Output directory in which to save the log file. The default is ``None``.
        ports : str
            Number of excitations expected. The default is ``None``.

        Returns
        -------
        type
            All the info in a list for later use

        """
        if name is None:
            name= self.design_name
        if outputdir is None:
            outputdir = self.project_path

        self._messenger.add_info_message("#### Design Validation Checks###")
        #
        # Routine outputs to the validation info to a log file in the project directory and also
        # returns the validation info to be used to update properties.xml file

        validation_ok = True

        #
        # Write an overall validation log file with all output from all checks
        # The design validation inside HFSS outputs to a separate log file which we merge into this overall file
        #
        val_list = []
        all_validate = outputdir + "\\all_validation.log"
        with open(all_validate, "w") as validation:

            # Desktop Messages
            msg = "Desktop Messages:"
            validation.writelines(msg + "\n")
            val_list.append(msg)
            msgs = self._desktop.GetMessages(name, "HFSSDesign1", 0)
            # need to check if design name is always this default name HFSSDesign1
            for msg in msgs:
                self._messenger.add_info_message(msg)
                # msg = msg.replace('"','')
                msg = msg.rstrip('\r\n')
                val_list.append(msg)
                validation.writelines(msg + "\n")

            # Run Design Validation and write out the lines to the logger

            ret = self._odesign.ValidateCircuit()
            msg = "Design Validation Messages:"
            validation.writelines(msg + "\n")
            val_list.append(msg)
            if ret == 0:
                msg = "**** ERRORS Present - please check and confirm"
                self._messenger.add_error_message(msg)
            else:
                msg = "**** Validation Completed Correctly"
                self._messenger.add_info_message(msg)

            # Find the Excitations and check or list them out
            msg = "Excitation Messages:"
            validation.writelines(msg + "\n")
            val_list.append(msg)
            numportsdefined = int(len(self.get_excitations_name))
            if ports is not None and ports != numportsdefined:
                msg = "**** Port Number Error! - Please check model"
                self._messenger.add_error_message(msg)
                validation.writelines(msg + "\n")
                val_list.append(msg)
                validation_ok = False
                # need to stop the simulation athis point
            else:
                msg1 = "Ports Requested: " + str(ports)
                msg2 = "Ports Defined: " + str(numportsdefined)
                self._messenger.add_info_message(msg1)
                validation.writelines(msg1 + "\n")
                val_list.append(msg1)
                self._messenger.add_info_message(msg2)
                validation.writelines(msg2 + "\n")
                val_list.append(msg2)

            excitation_names = self.get_excitations_name
            for excitation in excitation_names:
                msg = "Excitation name: " + str(excitation)
                self._messenger.add_info_message(msg)
                validation.writelines(msg + "\n")
                val_list.append(msg)
        validation.close()
        return val_list, validation_ok  # return all the info in a list for use later

    @aedt_exception_handler
    def create_scattering(self, plot_name="S Parameter Plot Nominal", sweep_name=None, port_names=None, port_excited=None, variations=None):
        """Create scattering Report
        

        Parameters
        ----------
        PlotName :
             The name of the plot. The default is ``"S Parameter Plot Nominal"``
        sweep_name : str
             The default is ``None``.
        PortNames : str
             The names of the port. The default is ``None``.
        PortExcited : str
             The default is ``None``.
        variations :
             The default is ``None``.

        Returns
        -------

        """
        Families = ["Freq:=", ["All"]]
        if variations:
            Families +=variations
        else:
            Families += self.get_nominal_variation()
        if not sweep_name:
            sweep_name = self.existing_analysis_sweeps[1]
        if not port_names:
            port_names = self.get_excitations_name
        if not port_excited:
            port_excited= port_names
        Trace = ["X Component:=", "Freq", "Y Component:=", ["dB(S(" + p + "," + q + "))" for p,q in zip(list(port_names), list(port_excited))]]
        solution_data = ""
        if self.solution_type == "DrivenModal":
            solution_data = "Modal Solution Data"
        elif self.solution_type == "DrivenTerminal":
            solution_data = "Terminal Solution Data"
        elif self.solution_type == "HFSS3DLayout":
            solution_data = "Standard"
        if solution_data != "":
            # run CreateReport function
            self.post.oreportsetup.CreateReport(
                plot_name,
                solution_data,
                "Rectangular Plot",
                sweep_name,
                ["Domain:=", "Sweep"],
                Families,
                Trace,
                [])
            return True
        else:
            return False

    @aedt_exception_handler
    def export_touchstone(self, solutionname, sweepname, filename, variation, variations_value):
        """Export the Touchstone file.
        
        Parameters
        ----------
        solutionname : str
            Name of the solution that has been solved.    
        sweepname : str
            Name of the sweep that has been solved.
        filename : str
            Full path of the output file.
        variation : list
            List of all parameter variations, such  as ``["$AmbientTemp", "$PowerIn"]``.
        variations_value : list
            List of all parameter variation values, such as ``["22cel", "100"]``.

        Returns
        -------

        """

        # normalize the save path
        if not filename:
            appendix = ""
            for v, vv in zip(variation, variations_value):
                appendix += "_" + v + vv.replace("\'", "")
            ext = ".S" + str(len(self.port_list)) + "p"
            filename = os.path.join(self.project_path, solutionname + "_" + sweepname + appendix + ext)
        else:
            filename = filename.replace("//", "/").replace("\\", "/")
        print("Exporting Touchstone " + filename)
        # array containing "SetupName:SolutionName" pairs (note that setup and solution are separated by a colon)
        SolutionSelectionArray = [solutionname + ":" + sweepname]
        # 2=tab delimited spreadsheet (.tab), 3= touchstone (.sNp), 4= CitiFile (.cit),
        # 7=Matlab (.m), 8=Terminal Z0 spreadsheet
        FileFormat = 3
        OutFile = filename  # full path of output file
        FreqsArray = ["all"]  # array containin the frequencies to export, use ["all"] for all frequencies
        DoRenorm = True  # perform renormalization before export
        RenormImped = 50  # Real impedance value in ohm, for renormalization
        DataType = "S"  # Type: "S", "Y", or "Z" matrix to export
        Pass = -1  # The pass to export. -1 = export all passes.
        ComplexFormat = 0  # 0=Magnitude/Phase, 1=Real/Immaginary, 2=dB/Phase
        DigitsPrecision = 15  # Touchstone number of digits precision
        IncludeGammaImpedance = True  # Include Gamma and Impedance in comments
        NonStandardExtensions = False  # Support for non-standard Touchstone extensions
        variation_str = ""
        for v, vv in zip(variation, variations_value):
            variation_str += v + "=" + vv + " "

        self.odesign.ExportNetworkData(variation_str, SolutionSelectionArray, FileFormat,
                                       OutFile, FreqsArray, DoRenorm, RenormImped, DataType, Pass, ComplexFormat,
                                       DigitsPrecision, IncludeGammaImpedance, IncludeGammaImpedance,
                                       NonStandardExtensions)
        return True

    @aedt_exception_handler
    def set_export_touchstone(self, activate, export_dir=""):
        """Set automatic export of the Touchstone file after the simulation is ``True``.

        Parameters
        ----------
        activate : bool
            Export after the simulation.

        Returns
        -------
        type
            True if operation succeeded

        """

        settings = []
        if activate:
            settings.append("NAME:options")
            settings.append("ExportAfterSolve:=")
            settings.append(True)
            settings.append("ExportDir:=")
            settings.append(export_dir)
        elif not activate:
            settings.append("NAME:options")
            settings.append("ExportAfterSolve:=")
            settings.append(False)
        self.odesign.DesignOptions(settings, 0)
        return True

    @aedt_exception_handler
    def create_frequency_sweep(self, setupname, unit, freqstart, freqstop, num_of_freq_points,
                               sweepname=None, sweeptype="interpolating",
                               interpolation_tol_percent=0.5, interpolation_max_solutions=250,
                               save_fields=True, save_rad_fields_only=False,
                               use_q3d_for_dc=False):
        """Create a frequency sweep.

        Parameters
        ----------
        setupname : str
            Name of the setup that is attached to the sweep.
        unit :
            Units such as ``"MHz"``, ``"GHz"``, and so on.
        freqstart :
            Starting frequency of the sweep.
        freqstop :
            Stopping frequency of the sweep.
        sweepname : str
            Name of the sweep. The default is ``None``.
        num_of_freq_points :
            Number of frequency points in the range.
        sweeptype : discrete
            Type of sweep. Choices are ``"Fast"``, ``"Interpolating"``, and ``"Discrete"``. The default is ``"Interpolating``.
        interpolation_max_solutions :
            Maximum number of solutions evaluated for the interpolation process. The default is ``250``.
        interpolation_tol_percent :
            Error tolerance threshold for the interpolation process. The default is ``0.5``.
        save_fields : bool
            Save the fields for a discrete sweep only. The default is ``True``.
        save_rad_fields_only : bool
            Save only the radiated fields if ``save_fields = True``. The default is ``False``.
        use_q3d_for_dc : bool
            Use Q3D to solve the DC point. The default is ``False``.

        Returns
        -------
        type
            Setup name if operation succeeded

        """
        if sweepname is None:
            sweepname = generate_unique_name("Sweep")
        interpolation = False
        if sweeptype == "interpolating":
            interpolation = True
            save_fields = False

        if not save_fields:
            save_rad_fields_only = False
        interpolation_tol = interpolation_tol_percent / 100.
        sweep = SweepString()
        sweep.add_sweep([freqstart, freqstop, num_of_freq_points], "linear_count", unit)
        # the add_sweep function supports all sweeps type e.g.
        #   sweep.add_sweep([10, 20, 1], "linear_step", "MHz")
        #   sweep.add_sweep([1, 1000, 100], "log_scale", "kHz")
        #   sweep.add_sweep([7,13,17,19,23], "single", unit)
        sweep_string = sweep.get_string()
        setup = self.get_setup(setupname)
        if sweepname in [name.name for name in setup.sweeps]:
            sweepname = generate_unique_name(sweepname)
        sweep= setup.add_sweep(sweepname=sweepname)
        sweep.change_range("LinearCount", freqstart,freqstop, num_of_freq_points )
        setup.props["GenerateSurfaceCurrent"]=save_fields
        setup.props["SaveRadFieldsOnly"] = save_rad_fields_only
        setup.props["FastSweep"] = interpolation
        setup.props["SAbsError"] = interpolation_tol
        setup.props["EnforcePassivity"] = interpolation
        setup.props["UseQ3DForDC"] = use_q3d_for_dc
        setup.props["MaxSolutions"] = interpolation_max_solutions
        setup.update()
        # arg = ["NAME:" + sweepname,
        #        [
        #         "NAME:Properties",
        #         "Enable:=", "true"
        #        ],
        #        [
        #         "NAME:Sweeps",
        #         "Variable:=", sweepname,
        #         "Data:=", sweep_string,
        #         "OffsetF1:=", False,
        #         "Synchronize:=", 0
        #        ],
        #        "GenerateSurfaceCurrent:=", save_fields,
        #        "SaveRadFieldsOnly:=", save_rad_fields_only,
        #        "FastSweep:=", interpolation,
        #        "ZoSelected:=", False,
        #        "SAbsError:=", interpolation_tol,
        #        "ZoPercentError:=", 1,
        #        "GenerateStateSpace:=", False,
        #        "EnforcePassivity:=", interpolation,
        #        "PassivityTolerance:=", 0.0001,
        #        "UseQ3DForDC:=", use_q3d_for_dc,
        #        "ResimulateDC:=", False,
        #        "MaxSolutions:=", interpolation_max_solutions,
        #        "InterpUseSMatrix:=", True,
        #        "InterpUsePortImpedance:=", True,
        #        "InterpUsePropConst:=", True,
        #        "InterpUseFullBasis:=", True,
        #        "CustomFrequencyString:=", "",
        #        "AllEntries:=", False,
        #        "AllDiagEntries:=", False,
        #        "AllOffDiagEntries:=", False,
        #        "MagMinThreshold:=", 0.01
        #        ]
        #
        # self.oanalysis.AddSweep(setupname, arg)
        # self.oanalysis_setup.AddSweep(setupname, arg)
        # self._messenger.add_debug_message("Sweep Setup created correctly")
        return setup

    @aedt_exception_handler
    def import_gds(self, gds_path, aedb_path=None, xml_path=None, set_as_active=True, close_active_project=False):
        """Import Gds into HFSS3DLayout and assign stackup from xml if present.

        Parameters
        ----------
        gds_path : str
            Full path to .gds file
        aedb_path : str, optional
            Full path to aedb file
        xml_path : str, optional
            path to stackup information. If not provided, the stackup will not be edited
        set_as_active : bool
            Set Gds as active project
        close_active_project : bool
            Close active project after loading the gds.
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        active_project = self.project_name
        project_name = os.path.basename(gds_path)[:-4]
        if not aedb_path:
            aedb_path = gds_path.replace('.gds', '.aedb')
        if os.path.exists(aedb_path):
            old_name = project_name
            project_name = generate_unique_name(project_name)
            aedb_path = gds_path.replace(old_name + '.gds', project_name + '.aedb')
            self.messenger.add_warning_message("aedb_exists. Renaming it to {}".format(project_name))

        oTool = self.odesktop.GetTool("ImportExport")
        oTool.ImportGDSII(gds_path, aedb_path, "", "")
        project = self.odesktop.SetActiveProject(project_name)
        oeditor = project.GetActiveDesign().SetActiveEditor("Layout")
        if xml_path:
            oeditor.ImportStackupXML(xml_path)
        if set_as_active:
            self.__init__(project_name)
        if close_active_project:
            self.odesktop.CloseProject(active_project)
        return True





import csv
import re

from ..generic.general_methods import aedt_exception_handler, generate_unique_name
from ..application.Analysis import Analysis
from ..modeler.Model3D import Modeler3D
from ..modules.MeshIcepak import IcepakMesh


class FieldAnalysisIcepak(Analysis, object):
    """AEDT_Icepak_FieldAnalysis
    Class for 3D Field Analysis Setup (Icepak)

    It is automatically initialized by Application call (like HFSS,
    Q3D...). Refer to Application function for inputs definition

    Parameters
    ----------

    Returns
    -------

    """
    def __init__(self, application, projectname, designname, solutiontype, setup_name=None,
                 specified_version=None, NG=False, AlwaysNew=False, release_on_exit=False):
        Analysis.__init__(self, application, projectname, designname, solutiontype, setup_name,
                          specified_version, NG, AlwaysNew, release_on_exit)
        self._modeler = Modeler3D(self)
        self._mesh = IcepakMesh(self)
        #self._post = PostProcessor(self)
        self.modeler.primitives.refresh()
    @property
    def modeler(self):
        """ """
        return self._modeler

    @property
    def mesh(self):
        """ """
        return self._mesh

    # @property
    # def post(self):
    #     return self._post

    @aedt_exception_handler
    def apply_icepak_settings(self, ambienttemp=20, gravityDir=5, perform_minimal_val=True, default_fluid="air",
                              default_solid="Al-Extruded", default_surface="Steel-oxidised-surface"):
        """Apply Icepak Default Design Settings like Ambient Temperature and Gravity.

        Parameters
        ----------
        ambienttemp :
            Ambient Temperature. It can be an integer or a parameter already created in AEDT (Default value = 20)
        gravityDir :
            Gravity direction index [0, 5] (Default value = 5)
        perform_minimal_val :
            True Perform Minimal validation. False perform Full Validation (Default value = True)
        default_fluid :
            default Fluid ."Air"
        default_solid :
            default Solid ."Al-Extruded"
        default_surface :
            default Surface ."Steel-oxidised-surface"

        Returns
        -------
        type
            True

        """
        if type(ambienttemp) is int:
            AmbientTemp = str(ambienttemp)+"cel"
        else:
            AmbientTemp = ambienttemp

        IceGravity = ["X", "Y", "Z"]
        GVPos = False
        if int(gravityDir) > 2:
            GVPos = True
        GVA = IceGravity[int(gravityDir) - 3]
        self.odesign.SetDesignSettings(
            ["NAME:Design Settings Data", "Perform Minimal validation:=", perform_minimal_val, "Default Fluid Material:=",
             default_fluid, "Default Solid Material:=", default_solid, "Default Surface Material:=", default_surface,
             "AmbientTemperature:=", AmbientTemp, "AmbientPressure:=", "0n_per_meter_sq",
             "AmbientRadiationTemperature:=", AmbientTemp,
             "Gravity Vector CS ID:=", 1, "Gravity Vector Axis:=", GVA, "Positive:=", GVPos],
            ["NAME:Model Validation Settings"])
        return True

    @aedt_exception_handler
    def get_property_value(self, objectname, property, type=None):
        """Get Design Property Value of specific object.

        Parameters
        ----------
        type :
            type of the property. "Boundary", "Excitation", "Setup", "Mesh" (Default value = None)
        objectname :
            param property:
        property :
            

        Returns
        -------
        type
            property value
            
            
            :Example:
            
            val = ipk.get_property_value('BoundarySetup:Source1', 'Total Power')

        """
        boundary = {"HFSS": "HfssTab", "Icepak": "Icepak", "Q3D": "Q3D", "Maxwell3D": "Maxwell3D"}
        excitation = {"HFSS": "HfssTab", "Icepak": "Icepak", "Q3D": "Q3D", "Maxwell3D": "Maxwell3D"}
        setup = {"HFSS": "HfssTab", "Icepak": "Icepak", "Q3D": "General", "Maxwell3D": "General"}
        mesh = {"HFSS": "MeshSetupTab", "Icepak": "Icepak", "Q3D": "Q3D", "Maxwell3D": "Maxwell3D"}
        all = {"HFSS": ["HfssTab", "MeshSetupTab"], "Icepak": ["Icepak"], "Q3D": ["Q3D", "General"], "Maxwell3D" : ["Maxwell3D", "General"]}
        if type == "Boundary":
            propserv = boundary[self._design_type]
            val = self.odesign.GetPropertyValue(propserv, objectname, property)
            return val

        elif type == "Setup":
            propserv = setup[self._design_type]
            val = self.odesign.GetPropertyValue(propserv, objectname, property)
            return val
        elif type == "Excitation":
            propserv = excitation[self._design_type]
            val = self.odesign.GetPropertyValue(propserv, objectname, property)
            return val

        elif type == "Mesh":
            propserv = mesh[self._design_type]
            val = self.odesign.GetPropertyValue(propserv, objectname, property)
            return val
        else:
            propservs = all[self._design_type]
            for propserv in propservs:
                properties = list(self.odesign.GetProperties(propserv, objectname))
                if property in properties:
                    val = self.odesign.GetPropertyValue(propserv, objectname, property)
                    return val
        return None

    @aedt_exception_handler
    def copy_solid_bodies_from(self, design, object_list=None, no_vacuum=True, no_pec=True):
        """Copy all the list of object from one design to active one

        Parameters
        ----------
        design :
            starting application object (examples hfss1= HFSS3DLayout)
        object_list :
            List of object to copy (Default value = None)
        no_vacuum : bool
            define if vacuum objects have to be copied (Default value = True)
        no_pec : bool
            define if pec objects have to be copied (Default value = True)

        Returns
        -------
        type
            True if succeeded

        """
        body_list = design.modeler.solid_bodies
        selection_list = []
        material_properties = design.modeler.primitives.objects
        for body in body_list:
            include_object = True
            if object_list:
                if body not in object_list:
                    include_object = False
            for key, val in material_properties.items():
                if val.name == body:
                    if no_vacuum and val.material_name == 'Vacuum':
                        include_object = False
                    if no_pec and val.material_name == 'pec':
                        include_object = False
            if include_object:
                selection_list.append(body)
        design.modeler.oeditor.Copy(["NAME:Selections", "Selections:=", ','.join(selection_list)])
        self.modeler.oeditor.Paste()
        self.modeler.primitives.refresh_all_ids()
        return True

    @aedt_exception_handler
    def assignmaterial(self, obj, mat):
        """The function assigns Material mat to object obj. If material mat is not present it will be created

        Parameters
        ----------
        obj : str, list
            list of objects to which assign materials
        mat : str
            material to assign

        Returns
        -------
        type
            True if succeeded | False if failed

        """
        mat = mat.lower()
        selections = self.modeler.convert_to_selections(obj)
        arg1 = ["NAME:Selections"]
        arg1.append("Selections:="), arg1.append(selections)
        arg2 = ["NAME:Attributes"]
        arg2.append("MaterialValue:="), arg2.append(chr(34) + mat + chr(34))
        if mat in self.materials.material_keys:
            Mat = self.materials.material_keys[mat]
            Mat.update()
            if Mat.is_dielectric():
                arg2.append("SolveInside:="), arg2.append(True)
            else:
                arg2.append("SolveInside:="), arg2.append(False)
            self.modeler.oeditor.AssignMaterial(arg1, arg2)
            self._messenger.add_info_message('Assign Material ' + mat + ' to object ' + selections)
            self.materials._aedmattolibrary(mat)
            return True
        elif self.materials.checkifmaterialexists(mat):
            Mat = self.materials.material_keys[mat]
            if Mat.is_dielectric():
                arg2.append("SolveInside:="), arg2.append(True)
            else:
                arg2.append("SolveInside:="), arg2.append(False)
            self.modeler.oeditor.AssignMaterial(arg1, arg2)
            self._messenger.add_info_message('Assign Material ' + mat + ' to object ' + selections)
            return True
        else:
            self._messenger.add_error_message("Material Does Not Exists")
            return False

    @aedt_exception_handler
    def _assign_property_to_mat(self,newmat, val, property):
        """

        Parameters
        ----------
        newmat :
            
        val :
            
        property :
            

        Returns
        -------

        """
        try:
            if "@" not in val:
                value = float(val)
                newmat.set_property_value(property, value)

            else:
                value_splitted = val.split(",")
                value_list = [
                    [float(re.findall("[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?", a)[0]) for a in
                     d.split("@")] for d in value_splitted]
                val0 = float(value_list[0][0])
                for el in value_list:
                    el.reverse()
                    el[1] = float(el[1])/val0
                newmat.set_property_value(property, val0)
                dataset = newmat.create_thermal_modifier(value_list)
                newmat.set_property_therm_modifier(property, dataset)
            return True
        except:
            return False

    @aedt_exception_handler
    def assignmaterial_from_sherlock_files(self, csv_component, csv_material):
        """Assign material to objects in design based on csv files obtained from Sherlock Tool

        Parameters
        ----------
        csv_component :
            csv containing component properties including material name
        csv_material :
            csv containing material properties

        Returns
        -------
        type
            Bool

        """
        with open(csv_material) as csvfile:
            csv_input = csv.reader(csvfile)
            material_header = next(csv_input)
            data = list(csv_input)
            k = 0
            material_data = {}
            for el in material_header:
                material_data[el] = [i[k] for i in data]
                k += 1
        with open(csv_component) as csvfile:
            csv_input = csv.reader(csvfile)
            component_header = next(csv_input)
            data = list(csv_input)
            k = 0
            component_data = {}
            for el in component_header:
                component_data[el] = [i[k] for i in data]
                k += 1
        all_objs = self.modeler.primitives.get_all_objects_names()
        i = 0
        for mat in material_data["Name"]:
            list_mat_obj = ["COMP_" + rd for rd, md in zip(component_data["Ref Des"], component_data["Material"]) if
                            md == mat]
            list_mat_obj += [rd for rd, md in zip(component_data["Ref Des"], component_data["Material"]) if
                            md == mat]
            list_mat_obj = [mo for mo in list_mat_obj if mo in all_objs]
            if list_mat_obj:
                if not self.materials.checkifmaterialexists(mat):
                    newmat = self.materials.creatematerial(mat)
                    if "Material Density" in material_data:
                        value = material_data["Material Density"][i]
                        self._assign_property_to_mat(newmat, value, newmat.PropName.MassDensity)
                    if "Thermal Conductivity" in material_data:
                        value = material_data["Thermal Conductivity"][i]
                        self._assign_property_to_mat(newmat, value, newmat.PropName.ThermalConductivity)
                    if "Material CTE" in material_data:
                        value = material_data["Material CTE"][i]
                        self._assign_property_to_mat(newmat, value, newmat.PropName.ThermalExpCoeff)
                    if "Poisson Ratio" in material_data:
                        value = material_data["Poisson Ratio"][i]
                        self._assign_property_to_mat(newmat, value, newmat.PropName.PoissonsRatio)
                    if "Elastic Modulus" in material_data:
                        value = material_data["Elastic Modulus"][i]
                        self._assign_property_to_mat(newmat, value, newmat.PropName.YoungsModulus)
                    newmat.update()
                self.assignmaterial(list_mat_obj, mat)
                for o in list_mat_obj:
                    self.modeler.primitives.objects[self.modeler.primitives.get_obj_id(o)].material_name = mat
            i += 1
            all_objs = [ao for ao in all_objs if ao not in list_mat_obj]
        return True

    @aedt_exception_handler
    def get_all_conductors_names(self):
        """Get all conductors in active design
        
        
        :return: objname list

        Parameters
        ----------

        Returns
        -------

        """
        cond = [i.lower() for i in list(self.materials.conductors)]
        obj_names = []
        for el in self.modeler.primitives.objects:
            if self.modeler.primitives.objects[el].material_name in cond:
                obj_names.append(self.modeler.primitives.get_obj_name(el))
        return obj_names

    @aedt_exception_handler
    def get_all_dielectrics_names(self):
        """Get all dielectrics in active design
        
        
        :return: objname list

        Parameters
        ----------

        Returns
        -------

        """
        diel = [i.lower() for i in list(self.materials.dielectrics)]
        obj_names = []
        for el in self.modeler.primitives.objects:
            if self.modeler.primitives.objects[el].material_name in diel:
                obj_names.append(self.modeler.primitives.get_obj_name(el))
        return obj_names



"""
This module contains the ``analysis`` class.

It includes common classes for file management and messaging and all 
calls to AEDT modules like the modeler, mesh, postprocessing, and setup.
"""
from __future__ import absolute_import
import os
import itertools
import threading
import shutil
from ..modeler.modeler_constants import CoordinateSystemPlane, CoordinateSystemAxis, GravityDirection, Plane
from ..modules.SolveSetup import Setup
from ..modules.SolutionType import SolutionType, SetupTypes
from ..modules.SetupTemplates import SetupKeys
from ..modules.DesignXPloration import DOESetups, DXSetups, ParametericsSetups, SensitivitySetups, StatisticalSetups, OptimizationSetups
from .Design import Design
from ..modules.MaterialLib import Materials
from ..generic.general_methods import aedt_exception_handler
from ..desktop import _pythonver
import sys
if "IronPython" in sys.version or ".NETFramework" in sys.version:
    from ..modules.PostProcessor import PostProcessor
else:
    from ..modules.AdvancedPostProcessing import PostProcessor


class Analysis(Design, object):
    """Analysis class.
    
    This is the main class, serving as a container for all common functions.
    
    It is automatically initialized by a call from an application, such as HFSS or Q3D. 
    See the application function for its parameter descriptions.

    Parameters
    ----------
    
    """

    def __init__(self, application, projectname, designname, solution_type, setup_name,
                 specified_version, NG, AlwaysNew, release_on_exit):
        self.setups = []
        Design.__init__(self, application, projectname, designname, solution_type,
                        specified_version, NG, AlwaysNew, release_on_exit)
        self.logger.info("Design Loaded")
        self._setup = None
        if setup_name:
            self.analysis_setup = setup_name
        self.solution_type = solution_type
        self._materials = Materials(self)
        self.logger.info("Materials Loaded")
        self._post = PostProcessor(self)
        self._available_variations = self.AvailableVariations(self)
        self.setups = [self.get_setup(setup_name) for setup_name in self.setup_names]
        self.opti_parametric = ParametericsSetups(self)
        self.opti_optimization = OptimizationSetups(self)
        self.opti_doe = DOESetups(self)
        self.opti_designxplorer = DXSetups(self)
        self.opti_sensitivity = SensitivitySetups(self)
        self.opti_statistical = StatisticalSetups(self)


    @property
    def materials(self):
        """Material Manager that is used to manage materials in the project.
        
        Parameters
        ----------

        Returns
        -------
        Materials
            Material Manager that is used to manage materials in the project.
           

        """
        return self._materials

    @property
    def Position(self):
        """Position of the object.
        
        Parameters
        ----------

        Returns
        -------
        type
            Position object.

        """
        return self.modeler.Position


    @property
    def available_variations(self):
        """Available Variation Object
        
        :return: Available Variation Object

        Parameters
        ----------

        Returns
        -------

        """

        return self._available_variations

    @property
    def CoordinateSystemAxis(self):
        """CoordinateSystemAxis Constant
        
        :return: Coordinate System Axis Constants Tuple (.X, .Y, .Z)

        Parameters
        ----------

        Returns
        -------

        """
        return CoordinateSystemAxis()

    @property
    def CoordinateSystemPlane(self):
        """CoordinateSystemPlane Constants
        
        :return: Coordinate System Plane Constants Tuple(.XY, .YZ, .XZ)

        Parameters
        ----------

        Returns
        -------

        """
        return CoordinateSystemPlane()

    @property
    def View(self):
        """Planes To check if redundant to CoordinateSystemPlane
        
        
        :return: Coordinate System Plane String Tuple("XY", "YZ", "XZ")

        Parameters
        ----------

        Returns
        -------

        """
        return Plane()

    @property
    def GravityDirection(self):
        """To check if redundant
        
        :return: Gravity Direction Tuple (XNeg, YNeg, ZNeg, XPos, YPos, ZPos)

        Parameters
        ----------

        Returns
        -------

        """
        return GravityDirection()

    @property
    def modeler(self):
        """:return: modeler object"""
        return self._modeler

    @property
    def mesh(self):
        """Mesh Object
        
        :return: mesh object

        Parameters
        ----------

        Returns
        -------

        """
        return self._mesh

    @property
    def post(self):
        """Post Processor Object
        
        :return: post processor object

        Parameters
        ----------

        Returns
        -------

        """
        return self._post

    @property
    def osolution(self):
        """osolution Object
        
        :return: Solutions Module object

        Parameters
        ----------

        Returns
        -------

        """
        return self.odesign.GetModule("Solutions")

    @property
    def oanalysis(self):
        """ """
        return self.odesign.GetModule("AnalysisSetup")

    @property
    def analysis_setup(self):
        """Analysis Setup
        
        
        :return: Active or first Analysis Setup Name

        Parameters
        ----------

        Returns
        -------

        """
        if self._setup:
            return self._setup
        elif self.existing_analysis_setups:
            return self.existing_analysis_setups[0]
        else:
            self._setup = None
            return self._setup

    @analysis_setup.setter
    def analysis_setup(self, setup_name):
        """

        Parameters
        ----------
        setup_name :
            

        Returns
        -------

        """
        setup_list = self.existing_analysis_setups
        if setup_list:
            assert setup_name in setup_list, "Invalid setup name {}".format(setup_name)
            self._setup = setup_name
        else:
            self._setup = setup_list[0]
        #return self._setup

    @property
    def existing_analysis_sweeps(self):
        """Existing Analysis Setup List
        
        
        :return: Return a list of all defined analysis setup names in the maxwell design.

        Parameters
        ----------

        Returns
        -------

        """
        setup_list = self.existing_analysis_setups
        sweep_list=[]
        for el in setup_list:
            if self.solution_type in SetupKeys.defaultAdaptive.keys():
                setuptype = SetupKeys.defaultAdaptive[self.solution_type]
                if setuptype:
                    sweep_list.append(el + " : " + setuptype)
            try:
                sweeps = list(self.oanalysis.GetSweeps(el))
            except:
                sweeps = []
            for sw in sweeps:
                sweep_list.append(el + " : " + sw)
        return sweep_list

    @property
    def nominal_adaptive(self):
        """

        Parameters
        ----------

        Returns
        -------
        type
            :return: str nominal adaptive setup

        """
        if len(self.existing_analysis_sweeps)>0:
            return self.existing_analysis_sweeps[0]
        else:
            return ""

    @property
    def nominal_sweep(self):
        """Return nominal sweep

        Returns
        -------
        str
            sweep is available it will return the lastadaptive
            
            :return: str nominal setup sweep if present
        """

        if len(self.existing_analysis_sweeps)>1:
            return self.existing_analysis_sweeps[1]
        else:
            return self.nominal_adaptive

    @property
    def existing_analysis_setups(self):
        """Existing Analysis Setup List
        
        
        :return: Return a list of all defined analysis setup names in the maxwell design.

        Parameters
        ----------

        Returns
        -------

        """
        setups = list(self.oanalysis.GetSetups())
        return setups

    @property
    def output_variables(self):
        """OutputVariable Object
        
        
        :return: Solutions OutputVariable object

        Parameters
        ----------

        Returns
        -------

        """
        oModule = self.odesign.GetModule("OutputVariable")
        return oModule.GetOutputVariables()

    @property
    def setup_names(self):
        """Setup Lists
        
        
        :return:  list of all defined analysis setup names in the design

        Parameters
        ----------

        Returns
        -------

        """
        setups = self.oanalysis.GetSetups()

        return setups


    @property
    def ooptimetrics(self):
        """Optimetrics Object
        
        :return: Solutions Optimetrics object

        Parameters
        ----------

        Returns
        -------

        """
        return self.odesign.GetModule("Optimetrics")

    @property
    def ooutput_variable(self):
        """OutputVariable Object
        
        :return: Solutions OutputVariable object

        Parameters
        ----------

        Returns
        -------

        """
        return self.odesign.GetModule("OutputVariable")

    @property
    def SimulationSetupTypes(self):
        """:return:List of all Simulation Setup Types categorized by Application"""
        return SetupTypes()


    @property
    def SolutionTypes(self):
        """:return:List of all Solution type categorized by Application"""
        return SolutionType()


    class AvailableVariations(object):

        def __init__(self, parent):
            """class Containing Available Variations

            Parameters
            ----------
            parent :
                parent object

            Returns
            -------

            """
            self._parent = parent

        @property
        def variables(self):
            """:return:List of independent variables"""
            return [i for i in self._parent.variable_manager.independent_variables]

        @aedt_exception_handler
        def variations(self, setup_sweep=None):
            """

            Parameters
            ----------
            setup_sweep :
                Setup and Sweep on which search variations (Default value = None)

            Returns
            -------
            type
                variation families list

            """
            if not setup_sweep:
                setup_sweep = self._parent.existing_analysis_sweeps[0]
            vs = self._parent.osolution.GetAvailableVariations(setup_sweep)
            families=[]
            for v in vs:
                variations = v.split(" ")
                family=[]
                for el in self.variables:
                    family.append(el + ":=")
                    i=0
                    while i<len(variations):
                        if variations[i][0:len(el)] == el:
                            family.append([variations[i][len(el)+2:-1]])
                        i += 1
                families.append(family)
            return families

        @property
        def nominal(self):
            """ """
            families = []
            for el in self.variables:
                families.append(el+":=")
                families.append(["Nominal"])
            return families

        @property
        def nominal_w_values(self):
            """ """
            families = []
            if self._parent.design_type == "HFSS 3D Layout Design":
                listvar = list(self._parent.odesign.GetVariables())
                for el in listvar:
                    families.append(el+":=")
                    families.append([self._parent.odesign.GetVariableValue(el)])
            else:
                variation = self._parent.odesign.GetNominalVariation()
                for el in self.variables:
                    families.append(el+":=")
                    families.append([self._parent.odesign.GetVariationVariableValue(variation, el)])
            return families

        @property
        def nominal_w_values_dict(self):
            """ """
            families = {}
            if self._parent.design_type == "HFSS 3D Layout Design":
                listvar = list(self._parent.odesign.GetVariables())
                for el in listvar:
                    families[el] = self._parent.odesign.GetVariableValue(el)
            else:
                variation = self._parent.odesign.GetNominalVariation()
                for el in self.variables:
                    families[el] = self._parent.odesign.GetVariationVariableValue(variation, el)
            return families

        @property
        def all(self):
            """ """
            families=[]
            for el in self.variables:
                families.append(el+":=")
                families.append(["All"])
            return families



    class AxisDir(object):
        """AxisDir class.
        
        This data class contains constants for the axis directions.
        """
        (XNeg, YNeg, ZNeg, XPos, YPos, ZPos) = range(0, 6)

    @aedt_exception_handler
    def get_setups(self):
        """

        Parameters
        ----------

        Returns
        -------
        list
            List of all setup names.

        """
        setups = self.oanalysis.GetSetups()
        return list(setups)

    @aedt_exception_handler
    def get_nominal_variation(self):
        """:return: nominal variation"""
        return self.available_variations.nominal

    @aedt_exception_handler
    def get_sweeps(self, name):
        """Retrieve all sweep for a setup.

        Parameters
        ----------
        name : str
            Name of the setup.
            
        Returns
        -------
        list
            List of all sweep names for the setup.

        """
        sweeps=[]

        sweeps = self.oanalysis.GetSweeps(name)

        return list(sweeps)

    @aedt_exception_handler
    def export_parametric_results(self, sweepname, filename, exportunits=True):
        """Export a list of all parametric variations solved to a CSV file for a sweep.

        Parameters
        ----------
        sweepname : str
            Name of the optimetrics sweep.
        filename : str
            Full path to the CSV file.
        exportunits : bool, optional
            Whether to export units with the value. The default is ``True`. When ``False``, 
            only the value is exported.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """

        self.ooptimetrics.ExportParametricResults(sweepname, filename, exportunits)
        return True


    @aedt_exception_handler
    def analyse_from_initial_mesh(self):
        """Revert the solution to the initial mesh and re-run the solve.
                
        Returns
        -------
        bool
           ``True`` when successful, ``False`` when failed.
        """
        self.oanalysis.RevertSetupToInitial(self._setup)
        self.analyse_nominal()
        return True


    @aedt_exception_handler
    def analyse_nominal(self):
        """Revert the solution to the initial mesh and re-run the solve.
       
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        self.odesign.Analyze(self.analysis_setup)
        return True


    @aedt_exception_handler
    def generate_unique_setup_name(self, setup_name=None):
        """Generate a new setup with an unique name.
        
        Parameters
        ----------
        setup_name : str, optional
            Name of the setup. The default is ``None``.

        Returns
        -------
        str
        Name of the setup.
        """
        if not setup_name:
            setup_name = "Setup"
        index = 2
        while setup_name in self.existing_analysis_setups:
            setup_name = setup_name + "_{}".format(index)
            index += 1
        return setup_name

    @aedt_exception_handler
    def create_setup(self, setupname="MySetupAuto", setuptype=None, props={}):
        """Create a new setup.

        Parameters
        ----------
        setupname : str, optional
            Name of the setup. The default is ``"MySetupAuto"``.
        setuptype : optional
            Type of the setup. The default is ``None``, which means that the
            default type is applied.
        props : dict, optional
            Dictionary of properties with values. The default is ``{}``.

        Returns
        -------
        Setup
            Setup object.

        """
        if setuptype is None:
            if self.design_type == "Icepak" and self.solution_type=="Transient":
                setuptype = SetupKeys.defaultSetups["TransientTemperatureAndFlow"]
            else:
                setuptype = SetupKeys.defaultSetups[self.solution_type]
        name = self.generate_unique_setup_name(setupname)
        setup = Setup(self, setuptype, name)
        setup.create()
        if props:
            for el in props:
                setup.props[el] = props[el]
            setup.update()
        self.analysis_setup = name
        self.setups.append(setup)
        return setup

    @aedt_exception_handler
    def edit_setup(self, setupname, properties_dict):
        """Edit the current setup.

        Parameters
        ----------
        setupname : str
            Name of the setup.
        properties_dict : dict 
            Dictionary containing the property to update with the value.

        Returns
        -------
        type
            Setup object.
        """
        setuptype = SetupKeys.defaultSetups[self.solution_type]
        setup = Setup(self, setuptype, setupname, isnewsetup=False)
        setup.update(properties_dict)
        self.analysis_setup = setupname
        return setup


    @aedt_exception_handler
    def get_setup(self, setupname):
        """Get the setup from the current design.

        Parameters
        ----------
        setupname : str
            Name of the setup.

        Returns
        -------
        type
            Setup object.
        """

        setuptype = SetupKeys.defaultSetups[self.solution_type]
        setup = Setup(self, setuptype, setupname, isnewsetup=False)
        if setup.props:
            self.analysis_setup = setupname
        return setup

    @aedt_exception_handler
    def create_output_variable(self, variable, expression):
        """Create or modify the output variable.

        Parameters
        ----------
        variable : str
            Name of the variable.
        expression :
            New value for the variable.

        Returns
        -------
        bool
           ``True`` when successful, ``False`` when failed.
        """
        oModule = self.odesign.GetModule("OutputVariable")
        if variable in self.output_variables:
            oModule.EditOutputVariable(variable, expression, variable, self.existing_analysis_sweeps[0], self.solution_type, [])
        else:
            oModule.CreateOutputVariable(variable, expression, self.existing_analysis_sweeps[0], self.solution_type, [])
        return True

    @aedt_exception_handler
    def get_output_variable(self, variable, solution_name=None, report_type_name=None):
        """Retrieve the value of the output variable.

        Parameters
        ----------
        variable : str
            Name of the variable.
        solution_name : str, optional
            Name of the solution. The default is ``None``.
        report_type_name : str, optional
            Name of the report type. The default is ``None``.

        Returns
        -------
        type
            Value of the output variable.
        """
        oModule = self.odesign.GetModule("OutputVariable")
        assert variable in self.output_variables, "Output Variable {} does not exist".format(variable)
        nominal_variation = self.odesign.GetNominalVariation()
        sol_type = self.solution_type
        value = oModule.GetOutputVariableValue(variable, nominal_variation, self.existing_analysis_sweeps[0], self.solution_type, [])
        return value

    @aedt_exception_handler
    def get_object_material_properties(self, object_list=None, prop_names=None):
        """Retrieve the material properties for a list of given objects and return them in a dictionary.
        
        This high-level function ignores objects with no defined material properties.

        Parameters
        ----------
        object_list : list, optional
            List of objects for which to get material_properties. The default is ``None``. 
            If ``None``, all objects are considered. 
        prop_names : str or list
            The property or list of properties to export.  The default is ``None``. 
            If ``None``, all properties are exported. 

        Returns
        -------
        dict
            Dictionary of objects with material properties.
        """
        if object_list:
            if not isinstance(object_list, list):
                object_list = [object_list]
        else:
            object_list = self.modeler.primitives.get_all_objects_names()

        if prop_names:
            if not isinstance(prop_names, list):
                prop_names = [prop_names]

        dict = {}
        for entry in object_list:
            mat_name = self.modeler.primitives[entry].material_name
            mat_props = self._materials[mat_name]
            if prop_names is None:
                dict[entry] = mat_props._props
            else:
                dict[entry] = {}
                for prop_name in prop_names:
                    dict[entry][prop_name] = mat_props._props[prop_name]
        return dict


    @aedt_exception_handler
    def analyze_setup(self, name):
        """Analyze a specific design setup.

        Parameters
        ----------
        name : str
            Name of the setup, which can be an optimetric setup or a simple setup.

        Returns
        -------
        bool
           ``True`` when successful, ``False`` when failed.
        """
        if name in self.existing_analysis_setups:
            self._messenger.add_info_message("Solving design setup {}".format(name))
            self.odesign.Analyze(name)
        else:
            try:
                self._messenger.add_info_message("Solving Optimetrics")
                self.ooptimetrics.SolveSetup(name)
            except:
                self._messenger.add_error_message("Setup Not found {}".format(name))
                return False
        return True

    @aedt_exception_handler
    def solve_in_batch(self, filename=None, machine="local", run_in_thread=False):
        """Analyze a design setup in batch mode. 
        
        .. note::
           To use this function, the AEDT project must be closed.

        Parameters
        ----------
        filename : str, optional
            Name of the setup. The default is ``None``, which means that the active project 
            is to be solved.
        machine : str, optional
            Name of the machine if remote.  The default is ``"local"``. 
        run_in_thread : bool, optional
            Whether the batch command is to be submitted as a thread. The default is
            ``False``. 

        Returns
        -------
         bool
           ``True`` when successful, ``False`` when failed.
        """
        if not filename:
            filename = self.project_file
            self.close_project()
        if machine == "local":
            # -Monitor option used as workaround for R2 BatchSolve not exiting properly at the end of the Batch job
            options = " -ng -BatchSolve -Monitor "
        else:
            options = " -ng -distribute -machinelist list=" + machine + " -Batchsolve "

        print("Batch Solve Options: " + options)
        if os.name == 'posix':
            batch_run = os.path.join(self.desktop_install_dir + "/ansysedt" + chr(34) + options + chr(
                34) + filename + chr(34))
        else:
            batch_run = chr(34) + self.desktop_install_dir + "/ansysedt.exe" + chr(34) + options + chr(
                34) + filename + chr(34)

        '''
        check for existing solution directory and delete if present so we
        dont have old .asol files etc
        '''

        print("Solving model in batch mode on " + machine)
        print("Batch Job command:" + batch_run)
        if run_in_thread:
            def thread_run():
                """ """
                os.system(batch_run)
            x = threading.Thread(target=thread_run)
            x.start()
        else:
            os.system(batch_run)
        print("Batch Job finished")
        return True


    @aedt_exception_handler
    def submit_job(self, clustername, aedt_full_exe_path=None, numnodes=1, numcores=32, wait_for_license=True, setting_file=None):
        """Submit a job to be solved on a cluster.

        Parameters
        ----------
        clustername : str
            Name of the cluster to which to submit the job.
        aedt_full_exe_path : str, optional
            Full path to the AEDT executable file. The default is ``None``, which uses
            ``"/clustername/AnsysEM/AnsysEM2x.x/Win64/ansysedt.exe"``.
        numnodes : int, optional
            Number of nodes. The default is ``1``.
        numcores : int, optional
            Number of cores. The default is ``32``.
        wait_for_license : bool, option
             Whether to wait for the license to be validated. The default is ``True``.
        setting_file : str, optional
            Name of the file to use as a template. The default value is ``None``.
        
        Returns
        -------
        type
            ID of the job.

        """
        project_file=self.project_file
        project_path = self.project_path
        if not aedt_full_exe_path:
            version = self.odesktop.GetVersion()[2:6]
            if os.path.exists(r"\\"+clustername+r"\AnsysEM\AnsysEM{}\Win64\ansysedt.exe".format(version)):
                aedt_full_exe_path = r"\\\\\\\\"+clustername+r"\\\\AnsysEM\\\\AnsysEM{}\\\\Win64\\\\ansysedt.exe".format(version)
            elif os.path.exists(r"\\"+clustername+r"\AnsysEM\AnsysEM{}\Linux64\ansysedt".format(version)):
                aedt_full_exe_path = r"\\\\\\\\"+clustername+r"\\\\AnsysEM\\\\AnsysEM{}\\\\Linux64\\\\ansysedt".format(version)
            else:
                self.messenger.add_error_message("Aedt Path doesn't exists. Please provide a full path")
                return False
        else:
            if not os.path.exists(aedt_full_exe_path):
                self.messenger.add_error_message("Aedt Path doesn't exists. Please provide a full path")
                return False
            aedt_full_exe_path.replace("\\", "\\\\")

        self.close_project()
        path_file = os.path.dirname(__file__)
        destination_reg = os.path.join(project_path, "Job_settings.areg")
        if not setting_file:
            setting_file = os.path.join(path_file, "..", "misc","Job_Settings.areg")
        shutil.copy(setting_file, destination_reg)

        f1 = open(destination_reg, 'w')
        with open(setting_file) as f:
            lines = f.readlines()
            for line in lines:
                if "\\	$begin" == line[:8]:
                    lin = "\\	$begin \\'{}\\'\\\n".format(clustername)
                    f1.write(lin)
                elif "\\	$end" == line[:6]:
                    lin = "\\	$end \\'{}\\'\\\n".format(clustername)
                    f1.write(lin)
                elif "NumCores" in line:
                    lin = "\\	\\	\\	\\	NumCores={}\\\n".format(numcores)
                    f1.write(lin)
                elif "NumNodes=1" in line:
                    lin = "\\	\\	\\	\\	NumNodes={}\\\n".format(numnodes)
                    f1.write(lin)
                elif "ProductPath" in line:
                    lin = "\\	\\	ProductPath =\\'{}\\'\\\n".format(aedt_full_exe_path)
                    f1.write(lin)
                elif "WaitForLicense" in line:
                    lin = "\\	\\	WaitForLicense={}\\\n".format(str(wait_for_license).lower())
                    f1.write(lin)
                else:
                    f1.write(line)
        f1.close()
        return self.odesktop.SubmitJob(os.path.join(project_path, "Job_settings.areg"), project_file)
